"""
Chain node VAE builders
"""
from domainlab.compos.vae.compos.decoder_concat_vec_reshape_conv_gated_conv import (
    DecoderConcatLatentFCReshapeConvGatedConv,
)
from domainlab.compos.vae.compos.encoder_xydt_elevator import XYDTEncoderArg
from domainlab.compos.vae.zoo_vae_builders_classif import NodeVAEBuilderArg


class NodeVAEBuilderImgTopic(NodeVAEBuilderArg):
    """NodeVAEBuilderImgTopic."""

    def is_myjob(self, request):
        """is_myjob.

        :param request:
        """
        self.args = request.args
        flag = True  # @FIXME
        self.config_img(flag, request)
        return flag

    def build_encoder(self, device, topic_dim):
        """build_encoder.

        :param device:
        :param topic_dim:
        """
        encoder = XYDTEncoderArg(
            device,
            topic_dim,
            self.zd_dim,
            self.zx_dim,
            self.zy_dim,
            self.i_c,
            self.i_h,
            self.i_w,
            args=self.args,
        )
        return encoder

    def build_decoder(self, topic_dim):
        """build_decoder.

        :param topic_dim:
        """
        decoder = DecoderConcatLatentFCReshapeConvGatedConv(
            z_dim=self.zd_dim + self.zx_dim + self.zy_dim + topic_dim,
            i_c=self.i_c,
            i_w=self.i_w,
            i_h=self.i_h,
        )
        return decoder



"""
Utilities for dataset
"""
import datetime

import torch
import torch.utils.data as data_utils
from PIL import Image
from torch.utils.data import Dataset
from torchvision.utils import save_image

from domainlab.utils.logger import Logger


def fun_img_path_loader_default(path):
    """
    https://discuss.pytorch.org/t/handling-rgba-images/88428/4
    """
    return Image.open(path).convert("RGB")


def mk_fun_label2onehot(dim):
    """
    function generator
    index to onehot
    """

    def fun_label2onehot(label):
        """
        :param label:
        """
        m_eye = torch.eye(dim)
        return m_eye[label]

    return fun_label2onehot


def plot_ds(dset, f_name, batchsize=32):
    """
    :param dset:
    :param f_name:
    :param batchsize: batch_size
    """
    loader_tr = data_utils.DataLoader(dset, batch_size=batchsize, shuffle=False)
    for _, (img, _, *_) in enumerate(loader_tr):
        nrow = min(img.size(0), 8)
        save_image(img.cpu(), f_name, nrow=nrow)
        break  # only one batch


def plot_ds_list(ds_list, f_name, batchsize=8, shuffle=False):
    """
    plot list of datasets, each datasets in one row
    :param ds_list:
    :param fname:
    :param batchsize:
    :param shuffle:
    """
    list_imgs = []
    for dset in ds_list:
        loader = data_utils.DataLoader(dset, batch_size=batchsize, shuffle=shuffle)
        for _, (img, _, *_) in enumerate(loader):
            list_imgs.append(img)
            break
    comparison = torch.cat(list_imgs)
    save_image(comparison.cpu(), f_name, nrow=batchsize)


class DsetInMemDecorator(Dataset):
    """
    fetch all items of a dataset into memory
    """

    def __init__(self, dset, name=None):
        """
        :param dset: x, y, *d
        :param name: name of dataset
        """
        self.dset = dset
        self.item_list = []
        logger = Logger.get_logger()
        if name is not None:
            logger.info(f"loading dset {name}")
        t_0 = datetime.datetime.now()
        for i in range(len(self.dset)):
            self.item_list.append(self.dset[i])
        t_1 = datetime.datetime.now()
        logger.info(f"loading dataset to memory taken: {t_1 - t_0}")

    def __getitem__(self, idx):
        """
        :param idx:
        """
        return self.item_list[idx]

    def __len__(self):
        return self.dset.__len__()



"""Chain of Responsibility"""

__author__ = "Xudong Sun"

import abc

from domainlab.utils.logger import Logger


class Request4Chain(metaclass=abc.ABCMeta):
    """
    define all available fields of request to ensure operation safety
    """

    @abc.abstractmethod
    def convert(self, obj):
        """
        Convert an heavy weight object into request object with pre-defined behavior
        """
        raise NotImplementedError


class AbstractChainNodeHandler(metaclass=abc.ABCMeta):
    """Chain of Responsibility:
    1. Make sure the chain can be constructed successfully even one node fails to
    initialize its designated service/business object so services from other nodes will still
    be available.
    2. To ensure this decoupling, avoid overriding self.__init__() (the initializer/constructor)
    of the handler by using multiple inheritance.
    e.g. inherit both AbstractChainNodeHandler and Factory Interface since doing so couples the
    Factory initializer with the AbstractChainNodeHandler():
    if the initializer/constructor of the Factory does not work, it will
    affect the whole chain.
    3. Instead, return service/business object in a member function of AbstractChainNodeHandler
    self.init_business(), this can result in redundant code but is safest.
    4. A sub-optimal but still acceptable solution is to use Multiple Inheritance (inherit
    AbstractChainNodeHandler and Factory interface) but **only** override the
    self.init_business(*kargs, **kwargs) method (with concrete input arguments)
    of the Chain Handler so the initializer/constructor of the Chain Handler
    will always work. Factory can be returned by calling
    ChainNode.init_business(*kargs, **kwargs).
    This can still be coupling since there might be some interface methods in Factory,
    once you change the parent class, some concrete factories has not implemented that,
    which will break the initalization of the chain.
    """

    def __init__(self, success_node=None):
        """__init__.
        :param success_node: successor chain node which implement the AbstractChainNodeHandler
        interface
        """
        self._task = None
        self._success_node = success_node
        self._parent_node = None
        self.next_model = None
        if success_node is not None:
            success_node.set_parent(self)

    def set_parent(self, parent):
        self._parent_node = parent
        parent._success_node = self

    @abc.abstractmethod
    def is_myjob(self, request):
        """is_myjob.
        :param request: subclass can override request object to be string or function
        :return True/False
        """
        raise NotImplementedError

    @abc.abstractmethod
    def init_business(self, *kargs):
        """init_business: initialize **and** return the heavy weight business object for
        doing the real job
        :param request: subclass can override request object to be string or function
        :return: the constructed service object
        """
        raise NotImplementedError

    def handle(self, request):
        """This method invoke self.is_myjob() to check which node in the chain should handle the
        request
        :param request: subclass can override request object to be string or function
        :return: light weight AbstractChainNodeHandler
        """
        if self.is_myjob(request):
            return self
        if self._success_node is not None:
            return self._success_node.handle(request)
        err_msg = "option " + str(request) + " does not exist"
        logger = Logger.get_logger()
        logger.info(err_msg)
        logger.info("available options are")
        self.print_options()
        raise NotImplementedError(err_msg)

    def print_options(self):
        logger = Logger.get_logger()
        logger.info(str(self.__class__.__name__))
        if self._parent_node is not None:
            self._parent_node.print_options()


class DummyBusiness:
    message = "dummy business"


class DummyChainNodeHandlerBeaver(AbstractChainNodeHandler):
    """Dummy class to show how to inherit from Chain of Responsibility"""

    def init_business(self, *kargs, **kwargs):
        return DummyBusiness()

    def is_myjob(self, request):
        """is_myjob.
        :param request: subclass can override request object to be string or function
        :return True/False
        """
        return True


class DummyChainNodeHandlerLazy(AbstractChainNodeHandler):
    """Dummy class to show how to inherit from Chain of Responsibility"""

    def init_business(self, *kargs, **kwargs):
        return DummyBusiness()

    def is_myjob(self, request):
        """is_myjob.
        :param request: subclass can override request object to be string or function
        :return True/False
        """
        return False



import torch.nn as nn

from domainlab.compos.vae.compos.encoder_domain_topic_img2topic import (
    EncoderImg2TopicDistri,
)
from domainlab.compos.vae.compos.encoder_domain_topic_img_topic2zd import (
    EncoderSandwichTopicImg2Zd,
)


class EncoderImg2TopicDirZd(nn.Module):
    """ """

    def __init__(self, i_c, i_h, i_w, num_topics, device, zd_dim, args):
        """__init__.

        :param i_c:
        :param i_h:
        :param i_w:
        :param num_topics:
        :param device:
        :param zd_dim:
        :param img_h_dim:
            - (img->h_img, topic->h_topic)-> q_zd,
              the dimension to concatenate with topic vector to infer z_d
            - img->img_h_dim->topic distribution
        """
        super().__init__()
        self.device = device
        self.zd_dim = zd_dim

        self.add_module(
            "net_img2topicdistri",
            EncoderImg2TopicDistri((i_c, i_h, i_w), num_topics, device, args),
        )

        # [topic, image] -> [h(topic), h(image)] -> [zd_mean, zd_scale]
        self.add_module(
            "imgtopic2zd",
            EncoderSandwichTopicImg2Zd(
                self.zd_dim,
                (i_c, i_h, i_w),
                num_topics,
                img_h_dim=num_topics,
                args=args,
            ),
        )

    def forward(self, img):
        """forward.
        :param img:
        """
        # image->h_image
        # h_image->alpha
        q_topic, topic_q = self.net_img2topicdistri(img)
        # [topic, image] -> [h(topic), h(image)] -> [zd_mean, zd_scale]
        q_zd, zd_q = self.imgtopic2zd(img, topic_q)
        return q_topic, topic_q, q_zd, zd_q



from domainlab.compos.a_nn_builder import AbstractFeatExtractNNBuilderChainNode
from domainlab.utils.u_import_net_module import (
    build_external_obj_net_module_feat_extract,
)


def mkNodeFeatExtractNNBuilderExternFromFile(arg_name_net_path):
    """
    for each algorithm, there might exist different feature extractors, e.g.
    for diva, there can be class feature extractor and domain feature
    extractor
    """

    class _LNodeFeatExtractNNBuilderExternFromFile(
        AbstractFeatExtractNNBuilderChainNode
    ):
        """LNodeFeatExtractNNBuilderExternFromFile.
        Local class to return
        """

        def init_business(
            self, dim_out, args, flag_pretrain, remove_last_layer, isize=None
        ):
            """
            initialize **and** return the heavy weight business object for
            doing the real job
            :param request: subclass can override request object to be
            string or function
            :return: the constructed service object
            """
            pyfile4net = getattr(args, arg_name_net_path)
            net = build_external_obj_net_module_feat_extract(
                pyfile4net, dim_out, remove_last_layer
            )
            return net

        def is_myjob(self, args):
            """is_myjob."""
            pyfile4net = getattr(args, arg_name_net_path)
            return pyfile4net is not None

    return _LNodeFeatExtractNNBuilderExternFromFile



"""
all available tasks for domainlab
"""

from domainlab.arg_parser import mk_parser_main
from domainlab.compos.pcr.request import RequestTask
from domainlab.tasks.task_mini_vlcs import addtask2chain
from domainlab.tasks.task_mnist_color import NodeTaskMNISTColor10
from domainlab.utils.logger import Logger
from domainlab.utils.u_import import import_path


class TaskChainNodeGetter(object):
    """
    1. Hardcoded chain
    3. Return selected node
    """

    def __init__(self, args):
        self.args = args
        tpath = args.tpath
        self.tpath = tpath
        self.request = RequestTask(args)()
        if tpath is not None:
            self.task_module = import_path(self.tpath)

    def __call__(self):
        """
        1. construct the chain, filter out responsible node, create heavy-weight business object
        2. hard code seems to be the best solution
        """
        chain = NodeTaskMNISTColor10(None)
        chain = addtask2chain(chain)

        if self.tpath is not None:
            node = self.task_module.get_task(self.args.task)
            chain.set_parent(node)
            chain = node
            if self.args.task is None:
                logger = Logger.get_logger()
                logger.info("")
                logger.info(
                    f"overriding args.task {self.args.task} " f"to {node.task_name}"
                )
                logger.info("")
                self.request = node.task_name  # @FIXME
        node = chain.handle(self.request)
        return node


def get_task(name=None):
    """
    get build in task from DomainLab
    """
    args = mk_parser_main
    parser = mk_parser_main()
    args = parser.parse_args("")
    args.task = name
    task_getter = TaskChainNodeGetter(args)
    return task_getter()



"""
chain of responsibility pattern for algorithm selection
"""
from domainlab.algos.builder_api_model import NodeAlgoBuilderAPIModel
from domainlab.algos.builder_dann import NodeAlgoBuilderDANN
from domainlab.algos.builder_diva import NodeAlgoBuilderDIVA
from domainlab.algos.builder_erm import NodeAlgoBuilderERM
from domainlab.algos.builder_hduva import NodeAlgoBuilderHDUVA
from domainlab.algos.builder_jigen1 import NodeAlgoBuilderJiGen
from domainlab.utils.u_import import import_path


class AlgoBuilderChainNodeGetter:
    """
    1. Hardcoded chain
    3. Return selected node
    """

    def __init__(self, model, apath):
        self.model = model
        self.apath = apath
        #
        self._list_str_model = model.split("_")
        self.model = self._list_str_model.pop(0)

    def register_external_node(self, chain):
        """
        if the user specify an external python file to implement the algorithm
        """
        if self.apath is None:
            return chain
        node_module = import_path(self.apath)
        node_fun = node_module.get_node_na()  # @FIXME: build_node API need
        newchain = node_fun(chain)
        return newchain

    def __call__(self):
        """
        1. construct the chain, filter out responsible node, create heavy-weight business object
        2. hard code seems to be the best solution
        """
        chain = NodeAlgoBuilderDIVA(None)
        chain = NodeAlgoBuilderERM(chain)
        chain = NodeAlgoBuilderDANN(chain)
        chain = NodeAlgoBuilderJiGen(chain)
        chain = NodeAlgoBuilderHDUVA(chain)
        chain = NodeAlgoBuilderAPIModel(chain)
        chain = self.register_external_node(chain)
        node = chain.handle(self.model)
        head = node
        while self._list_str_model:
            self.model = self._list_str_model.pop(0)
            node2decorate = self.__call__()
            head.extend(node2decorate)
            head = node2decorate
        return node



import os
from importlib.machinery import SourceFileLoader


def import_path(path):
    path = os.path.expanduser(path)
    if path[0] == "/":
        full_path = path
    else:
        dir_path = os.getcwd()
        full_path = os.path.join(dir_path, path)
    return SourceFileLoader(fullname=full_path, path=full_path).load_module()



"""
upon a task, if Jigen is chosen as the model, then task's dataset has to be decorated with image tile permutation
note that task's dataset already include standard image
transformations like random croped resized, or flip, and normalization. See also the JiGen paper's implementation here:
https://github.com/fmcarlucci/JigenDG/blob/master/data/JigsawLoader.py
"""
import os

import numpy as np
import torch
import torchvision
from torch.utils import data as torchdata
from torchvision import transforms


class WrapDsetPatches(torchdata.Dataset):
    """
    given dataset of images, return permuations of tiles of images re-weaved
    """

    def __init__(
        self,
        dataset,
        num_perms2classify,
        prob_no_perm,
        grid_len,
        ppath=None,
        flag_do_not_weave_tiles=False,
    ):
        """
        :param prob_no_perm: probability of no permutation: permutation will change the image, so
        the class label classifier will behave very differently compared to no permutation
        """
        if ppath is None and grid_len != 3:
            raise RuntimeError(
                "please provide npy file of numpy array with each row \
                               being a permutation of the number of tiles, currently \
                               we only support grid length 3"
            )
        self.dataset = dataset
        self._to_tensor = transforms.Compose([transforms.ToTensor()])
        self.arr1perm_per_row = self.__retrieve_permutations(num_perms2classify, ppath)
        # for 3*3 tiles, there are 9*8*7*6*5*...*1 >> 100,
        # we load from disk instead only 100 permutations
        # each row of the loaded array is a permutation of the 3*3 tile
        # of the original image
        self.grid_len = grid_len
        # break the image into 3*3 tiles
        self.prob_no_perm = prob_no_perm
        if flag_do_not_weave_tiles:
            self.fun_weave_imgs = lambda x: x
        else:

            def make_grid(img):
                """
                sew tiles together to be an image
                """
                return torchvision.utils.make_grid(img, nrow=self.grid_len, padding=0)

            self.fun_weave_imgs = make_grid

    def get_tile(self, img, ind_tile):
        """
        assume a square image?
        """
        img_height = img.shape[-1]
        # @FIXME: use a better way to decide the image size
        num_tiles = float(img_height) / self.grid_len
        num_tiles = float(int(num_tiles)) + 1
        # @FIXME: extra line to ensure num_tiles=75 instead of sometimes 74
        # so torch.stack can fail in original data,
        # num_tiles = float(img.size[0]) / self.grid_len = 225/3 = 75.0
        # is an integer, but this can not be true for other cases
        ind_vertical = int(ind_tile / self.grid_len)
        ind_horizontal = ind_tile % self.grid_len
        functor_tr = transforms.ToPILImage()
        img_pil = functor_tr(img)
        # PIL.crop((left, top, right, bottom))
        # get rectangular region from box  of [left, upper, right, lower]
        tile = img_pil.crop(
            [
                ind_horizontal * num_tiles,
                ind_vertical * num_tiles,
                (ind_horizontal + 1) * num_tiles,
                (ind_vertical + 1) * num_tiles,
            ]
        )
        tile = self._to_tensor(tile)
        return tile

    def __getitem__(self, index):
        # image transformation from self.dataset happens here:
        img, label, *domain = self.dataset.__getitem__(index)
        # now img has been transformed (including normalization)
        original_size = img.shape[-2:]
        if domain:
            dlabel = domain[0]
        else:
            dlabel = None
        num_grids = self.grid_len**2
        # divide image into grid_len^2 tiles
        list_tiles = [None] * num_grids
        # list of length num_grids of image tiles
        for ind_tile in range(num_grids):
            list_tiles[ind_tile] = self.get_tile(img, ind_tile)  # populate tile list
        ind_which_perm = np.random.randint(self.arr1perm_per_row.shape[0] + 1)
        # +1 in line above is for when image is not permutated, which
        # also need to be classified corrected by the permutation classifier
        # let len(self.arr1perm_per_row)=31
        # so ind_which_perm is a random number in [0, 31] in total 31+1 classes
        # ind_which_perm is basically the row index to choose
        # from self.arr1perm_per_row which is a matrix of 31*9
        # where 9=3*3 is the number of tiles the image is broken into
        if self.prob_no_perm > 0:  # probability of no permutation of tiles
            # note that this "if" block is not redundant: permutation will change the image
            # thus change the behavior of the class label classifier, if self.prob_no_perm=1.0
            # then the algorithm will behave similarly to erm, though not completely same
            # FIXME: what hyperparameters one could set to let jigen=erm?
            if self.prob_no_perm > np.random.rand():
                ind_which_perm = 0
        # ind_which_perm = 0 means no permutation, the classifier need to
        # judge if the image has not been permutated as well
        list_reordered_tiles = None
        if ind_which_perm == 0:
            list_reordered_tiles = list_tiles  # no permutation of images
        else:  # default
            perm_chosen = self.arr1perm_per_row[ind_which_perm - 1]
            list_reordered_tiles = [
                list_tiles[perm_chosen[ind_tile]] for ind_tile in range(num_grids)
            ]
        stacked_tiles = torch.stack(list_reordered_tiles, 0)
        # NOTE: stacked_tiles will be [9, 3, 30, 30], which will be weaved to
        # be a whole image again by self.fun_weave_imgs
        # NOTE: ind_which_perm = 0 means no permutation, the classifier need to
        # judge if the image has not been permutated as well
        re_tiled_img = self.fun_weave_imgs(stacked_tiles)
        img_re_tiled_re_shaped = torchvision.transforms.RandomResizedCrop(
            original_size
        )(re_tiled_img)
        return img_re_tiled_re_shaped, label, dlabel, int(ind_which_perm)
        # ind_which_perm is the ground truth for the permutation index

    def __len__(self):
        return self.dataset.__len__()

    def __retrieve_permutations(self, num_perms_as_classes, ppath=None):
        """
        for 9 tiles which partition the image, we have num_perms_as_classes
        number of different permutations of the tiles, the classifier will
        classify the re-tile-ordered image permutation it come from.
        """
        # @FIXME: this assumes always a relative path
        mdir = os.path.dirname(os.path.realpath(__file__))
        if ppath is None:
            ppath = f"zdata/patches_permutation4jigsaw/permutations_{num_perms_as_classes}.npy"
        mpath = os.path.join(mdir, "..", ppath)
        arr_permutation_rows = np.load(mpath)
        # from range [1,9] to [0,8] since python array start with 0
        if arr_permutation_rows.min() == 1:
            arr_permutation_rows = arr_permutation_rows - 1
        return arr_permutation_rows



"""
Use dictionaries to create train and test domain split
"""
from torch.utils.data.dataset import ConcatDataset

from domainlab.tasks.a_task import NodeTaskDG
from domainlab.tasks.utils_task import DsetDomainVecDecorator, mk_loader, mk_onehot


class NodeTaskDict(NodeTaskDG):
    """
    Use dictionaries to create train and test domain split
    """

    def get_dset_by_domain(self, args, na_domain, split=False):
        """
        each domain correspond to one dataset, must be implemented by child class
        """
        raise NotImplementedError  # it is safe for each subclass to implement this

    def decorate_dset(self, model, args):
        """
        dispatch re-organization of data flow to model
        """

    def init_business(self, args, trainer=None):
        """
        create a dictionary of datasets
        """
        list_domain_tr, list_domain_te = self.get_list_domains_tr_te(
            args.tr_d, args.te_d
        )
        self.dict_dset_tr = {}
        self.dict_dset_val = {}
        dim_d = len(list_domain_tr)
        for ind_domain_dummy, na_domain in enumerate(list_domain_tr):
            dset_tr, dset_val = self.get_dset_by_domain(
                args, na_domain, split=args.split
            )
            vec_domain = mk_onehot(dim_d, ind_domain_dummy)  # for diva, dann
            ddset_tr = DsetDomainVecDecorator(dset_tr, vec_domain, na_domain)
            ddset_val = DsetDomainVecDecorator(dset_val, vec_domain, na_domain)
            if trainer is not None and hasattr(trainer, "dset_decoration_args_algo"):
                ddset_tr = trainer.dset_decoration_args_algo(args, ddset_tr)
                ddset_val = trainer.dset_decoration_args_algo(args, ddset_val)
            if (
                trainer is not None
                and trainer.model is not None
                and hasattr(trainer.model, "dset_decoration_args_algo")
            ):
                ddset_tr = trainer.model.dset_decoration_args_algo(args, ddset_tr)
                ddset_val = trainer.model.dset_decoration_args_algo(args, ddset_val)
            self.dict_dset_tr.update({na_domain: ddset_tr})
            self.dict_dset_val.update({na_domain: ddset_val})
        ddset_mix = ConcatDataset(tuple(self.dict_dset_tr.values()))
        self._loader_tr = mk_loader(ddset_mix, args.bs)

        ddset_mix_val = ConcatDataset(tuple(self.dict_dset_val.values()))
        self._loader_val = mk_loader(
            ddset_mix_val, args.bs, shuffle=False, drop_last=False
        )

        self.dict_dset_te = {}
        # No need to have domain Label for test
        for na_domain in list_domain_te:
            dset_te, *_ = self.get_dset_by_domain(args, na_domain, split=False)
            # NOTE: since get_dset_by_domain always return two datasets,
            # train and validation, this is not needed in test domain
            self.dict_dset_te.update({na_domain: dset_te})
        dset_te = ConcatDataset(tuple(self.dict_dset_te.values()))
        self._loader_te = mk_loader(dset_te, args.bs, shuffle=False, drop_last=False)



from domainlab.compos.builder_nn_alex import mkNodeFeatExtractNNBuilderNameAlex
from domainlab.compos.builder_nn_conv_bn_relu_2 import (
    mkNodeFeatExtractNNBuilderNameConvBnRelu2,
)
from domainlab.compos.builder_nn_external_from_file import (
    mkNodeFeatExtractNNBuilderExternFromFile,
)


class FeatExtractNNBuilderChainNodeGetter(object):
    """
    1. Hardcoded chain
    3. Return selected node
    """

    def __init__(self, args, arg_name_of_net, arg_path_of_net):
        """__init__.
        :param args: command line arguments
        :param arg_name_of_net: args.npath to specify
        where to get the external architecture for example
        """
        self.request = args
        self.arg_name_of_net = arg_name_of_net
        self.arg_path_of_net = arg_path_of_net

    def __call__(self):
        """
        1. construct the chain, filter out responsible node,
        create heavy-weight business object
        2. hard code seems to be the best solution
        """
        chain = mkNodeFeatExtractNNBuilderNameConvBnRelu2(
            self.arg_name_of_net, arg_val="conv_bn_pool_2", conv_stride=1
        )(None)
        chain = mkNodeFeatExtractNNBuilderNameConvBnRelu2(
            arg_name4net="nname_dom", arg_val="conv_bn_pool_2", conv_stride=1
        )(chain)
        chain = mkNodeFeatExtractNNBuilderNameAlex(self.arg_name_of_net, "alexnet")(
            chain
        )
        chain = mkNodeFeatExtractNNBuilderExternFromFile(self.arg_path_of_net)(chain)
        node = chain.handle(self.request)
        return node



def add_args2parser_vae(parser):
    parser.add_argument(
        "--zd_dim", type=int, default=64, help="diva: size of latent space for domain"
    )
    parser.add_argument(
        "--zx_dim",
        type=int,
        default=0,
        help="diva: size of latent space for unobserved",
    )
    parser.add_argument(
        "--zy_dim",
        type=int,
        default=64,
        help="diva, hduva: size of latent space for class",
    )
    #  HDUVA
    parser.add_argument(
        "--topic_dim", type=int, default=3, help="hduva: number of topics"
    )

    parser.add_argument(
        "--nname_encoder_x2topic_h",
        type=str,
        default=None,
        help="hduva: network from image to topic distribution",
    )

    parser.add_argument(
        "--npath_encoder_x2topic_h",
        type=str,
        default=None,
        help="hduva: network from image to topic distribution",
    )

    parser.add_argument(
        "--nname_encoder_sandwich_x2h4zd",
        type=str,
        default=None,
        help="hduva: network from image and topic to zd",
    )
    parser.add_argument(
        "--npath_encoder_sandwich_x2h4zd",
        type=str,
        default=None,
        help="hduva: network from image and topic to zd",
    )

    # ERM, ELBO
    parser.add_argument(
        "--gamma_y",
        type=float,
        default=None,
        help="diva, hduva: multiplier for y classifier",
    )
    parser.add_argument(
        "--gamma_d",
        type=float,
        default=None,
        help="diva: multiplier for d classifier from zd",
    )

    # Beta VAE part
    parser.add_argument(
        "--beta_t", type=float, default=1.0, help="hduva: multiplier for KL topic"
    )
    parser.add_argument(
        "--beta_d", type=float, default=1.0, help="diva: multiplier for KL d"
    )
    parser.add_argument(
        "--beta_x", type=float, default=1.0, help="diva: multiplier for KL x"
    )
    parser.add_argument(
        "--beta_y", type=float, default=1.0, help="diva, hduva: multiplier for KL y"
    )
    return parser



"""
Builder pattern to build different component for experiment with DIVA
"""
from domainlab.algos.a_algo_builder import NodeAlgoBuilder
from domainlab.algos.msels.c_msel_oracle import MSelOracleVisitor
from domainlab.algos.msels.c_msel_val import MSelValPerf
from domainlab.algos.observers.b_obvisitor import ObVisitor
from domainlab.algos.observers.c_obvisitor_cleanup import ObVisitorCleanUp
from domainlab.algos.observers.c_obvisitor_gen import ObVisitorGen
from domainlab.algos.trainers.zoo_trainer import TrainerChainNodeGetter
from domainlab.compos.pcr.request import RequestVAEBuilderCHW
from domainlab.compos.vae.utils_request_chain_builder import VAEChainNodeGetter
from domainlab.models.model_diva import mk_diva
from domainlab.utils.utils_cuda import get_device


class NodeAlgoBuilderDIVA(NodeAlgoBuilder):
    """
    Builder pattern to build different component for experiment with DIVA
    """

    def get_trainer(self, args):
        """
        chain of responsibility pattern for fetching trainer from dictionary
        """
        trainer = TrainerChainNodeGetter(args.trainer)(default="hyperscheduler")
        return trainer

    def init_business(self, exp):
        """
        return trainer, model, observer
        """
        task = exp.task
        args = exp.args
        request = RequestVAEBuilderCHW(task.isize.c, task.isize.h, task.isize.w, args)
        node = VAEChainNodeGetter(request)()
        task.get_list_domains_tr_te(args.tr_d, args.te_d)
        model = mk_diva()(
            node,
            zd_dim=args.zd_dim,
            zy_dim=args.zy_dim,
            zx_dim=args.zx_dim,
            list_str_y=task.list_str_y,
            list_d_tr=task.list_domain_tr,
            gamma_d=args.gamma_d,
            gamma_y=args.gamma_y,
            beta_x=args.beta_x,
            beta_y=args.beta_y,
            beta_d=args.beta_d,
        )
        device = get_device(args)
        model_sel = MSelOracleVisitor(MSelValPerf(max_es=args.es))
        if not args.gen:
            observer = ObVisitor(model_sel)
        else:
            observer = ObVisitorGen(model_sel)

        observer = ObVisitorCleanUp(observer)
        trainer = self.get_trainer(args)
        model = self.init_next_model(model, exp)
        return trainer, model, observer, device



"""
Color MNIST with single color
"""
import abc
import os
import struct

import numpy as np
from PIL import Image
from torch.utils.data import Dataset
from torchvision import datasets, transforms

from domainlab.dsets.utils_data import mk_fun_label2onehot
from domainlab.utils.utils_class import store_args


class ADsetMNISTColorRGBSolo(Dataset, metaclass=abc.ABCMeta):
    """
    Color MNIST with single color
    1. nominal domains: color palettes/range/spectrum
    2. subdomains: color(foreground, background)
    3. structure: each subdomain contains a combination of
    foreground+background color
    """

    @abc.abstractmethod
    def get_foreground_color(self, ind):
        raise NotImplementedError

    @abc.abstractmethod
    def get_background_color(self, ind):
        raise NotImplementedError

    @abc.abstractmethod
    def get_num_colors(self):
        raise NotImplementedError

    @store_args
    def __init__(
        self,
        ind_color,
        path="zoutput",
        subset_step=100,
        color_scheme="both",
        label_transform=mk_fun_label2onehot(10),
        list_transforms=None,
        raw_split="train",
        flag_rand_color=False,
    ):
        """
        :param ind_color: index of a color palette
        :param path: disk storage directory
        :param color_scheme:
            num(paint according to number),
            back(only paint background)
            both (background and foreground)
        :param list_transforms: torch transformations
        :param raw_split: default use the training part of mnist
        :param flag_rand_color: flag if to randomly paint each image
        (depreciated)
        :param label_transform:  e.g. index to one hot vector
        """
        dpath = os.path.normpath(path)
        flag_train = True
        if raw_split != "train":
            flag_train = False
        dataset = datasets.MNIST(
            root=dpath, train=flag_train, download=True, transform=transforms.ToTensor()
        )

        if color_scheme not in ["num", "back", "both"]:
            raise ValueError("color must be either 'num', 'back' or 'both")
        raw_path = os.path.dirname(dataset.raw_folder)
        self._collect_imgs_labels(raw_path, raw_split)
        inds_subset = list(range(0, len(dataset), subset_step))
        self.images = self.images[inds_subset, ::]
        self.labels = self.labels[inds_subset]
        self._color_imgs_onehot_labels()

    def _collect_imgs_labels(self, path, raw_split):
        """
        :param path:
        :param raw_split:
        """
        if raw_split == "train":
            fimages = os.path.join(path, "raw", "train-images-idx3-ubyte")
            flabels = os.path.join(path, "raw", "train-labels-idx1-ubyte")
        else:
            fimages = os.path.join(path, "raw", "t10k-images-idx3-ubyte")
            flabels = os.path.join(path, "raw", "t10k-labels-idx1-ubyte")

        # Load images
        with open(fimages, "rb") as f_h:
            _, _, rows, cols = struct.unpack(">IIII", f_h.read(16))
            self.images = np.fromfile(f_h, dtype=np.uint8).reshape(-1, rows, cols)

        # Load labels
        with open(flabels, "rb") as f_h:
            struct.unpack(">II", f_h.read(8))
            self.labels = np.fromfile(f_h, dtype=np.int8)
        self.images = np.tile(self.images[:, :, :, np.newaxis], 3)

    def __len__(self):
        return len(self.images)

    def _op_color_img(self, image):
        """
        transforms raw image into colored version
        """
        # randomcolor is a flag orthogonal to num-back-both
        if self.flag_rand_color:
            c_f = self.get_foreground_color(np.random.randint(0, self.get_num_colors()))
            c_b = 0
            if self.color_scheme == "both":
                count = 0
                while True:
                    c_b = self.get_background_color(
                        np.random.randint(0, self.get_num_colors())
                    )
                    if c_b != c_f and count < 10:
                        # exit loop if background color
                        # is not equal to foreground
                        break
        else:
            if self.color_scheme == "num":
                # domain and class label has perfect mutual information:
                # assign color
                # according to their class (0,10)
                c_f = self.get_foreground_color(self.ind_color)
                c_b = np.array([0] * 3)
            elif self.color_scheme == "back":  # only paint background
                c_f = np.array([0] * 3)
                c_b = self.get_background_color(self.ind_color)

            else:  # paint both background and foreground
                c_f = self.get_foreground_color(self.ind_color)
                c_b = self.get_background_color(self.ind_color)
        image[:, :, 0] = (
            image[:, :, 0] / 255 * c_f[0] + (255 - image[:, :, 0]) / 255 * c_b[0]
        )
        image[:, :, 1] = (
            image[:, :, 1] / 255 * c_f[1] + (255 - image[:, :, 1]) / 255 * c_b[1]
        )
        image[:, :, 2] = (
            image[:, :, 2] / 255 * c_f[2] + (255 - image[:, :, 2]) / 255 * c_b[2]
        )
        return image

    def _color_imgs_onehot_labels(self):
        """
        finish all time consuming operations before torch.dataset.__getitem__
        """
        for i in range(self.images.shape[0]):  # 60k*28*28*3
            img = self.images[i]  # size: 28*28*3 instead of 3*28*28
            self.images[i] = self._op_color_img(img)

    def __getitem__(self, idx):
        image = self.images[idx]  # range of pixel: [0,255]
        label = self.labels[idx]
        if self.label_transform is not None:
            label = self.label_transform(label)
        image = Image.fromarray(image)  # numpy array 28*28*3 -> 3*28*28
        if self.list_transforms is not None:
            for trans in self.list_transforms:
                image = trans(image)
        image = transforms.ToTensor()(image)  # range of pixel [0,1]
        return image, label



"""
Samples the hyperparameters according to a benchmark configuration file.

# Structure of this file:
- Class Hyperparameter
# Inherited Classes
# Functions to sample hyper-parameters and log into csv file
"""
import copy
import json
import os
from ast import literal_eval  # literal_eval can safe evaluate python expression
from pydoc import locate
from typing import List

import numpy as np
import pandas as pd

from domainlab.utils.get_git_tag import get_git_tag
from domainlab.utils.logger import Logger

G_MODEL_NA = "model"
G_METHOD_NA = "method"


class Hyperparameter:
    """
    Represents a hyperparameter.
    The datatype of .val is int if step and p1 is integer valued,
    else float.

    p1: min or mean
    p2: max or scale
    reference: None or name of referenced hyperparameter
    """

    def __init__(self, name: str):
        self.name = name
        self.val = 0

    def _ensure_step(self):
        """Make sure that the hyperparameter sticks to the discrete grid"""
        raise NotImplementedError

    def sample(self):
        """Sample this parameter, respecting properties"""
        raise NotImplementedError

    def get_val(self):
        """Returns the current value of the hyperparameter"""
        return self.val

    def datatype(self):
        """
        Returns the datatype of this parameter.
        This does not apply for references.
        """
        raise NotImplementedError


class SampledHyperparameter(Hyperparameter):
    """
    A numeric hyperparameter that shall be sampled
    """

    def __init__(self, name: str, config: dict):
        super().__init__(name)
        self.step = config.get("step", 0)
        try:
            self.distribution = config["distribution"]
            if self.distribution in {"uniform", "loguniform"}:
                self.p_1 = config["min"]
                self.p_2 = config["max"]
            elif self.distribution in {"normal", "lognormal"}:
                self.p_1 = config["mean"]
                self.p_2 = config["std"]
            else:
                raise RuntimeError(
                    f"Unsupported distribution type: {self.distribution}."
                )
        except KeyError as ex:
            raise RuntimeError(f"Missing required key for parameter {name}.") from ex

        self.p_1 = float(self.p_1)
        self.p_2 = float(self.p_2)

    def _ensure_step(self):
        """Make sure that the hyperparameter sticks to the discrete grid"""
        if self.step == 0:
            return  # continous parameter

        # round to next discrete value.
        # p_1 is the lower bound of the hyper-parameter range, p_2 the upper bound
        off = (self.val - self.p_1) % self.step
        # $off$ is always smaller than step, depending on whether $off$ falls on the left half
        # or right half of [0, step], move the hyper-parameter to the boundary so that
        # updated hyper-parameter % step = 0
        if off < self.step / 2:
            self.val -= off
        else:
            self.val += self.step - off
        # ensure correct datatype
        if self.datatype() == int:
            self.val = self.datatype()(np.round(self.val))

    def sample(self):
        """Sample this parameter, respecting properties"""
        if self.distribution == "uniform":
            self.val = np.random.uniform(self.p_1, self.p_2)
        elif self.distribution == "loguniform":
            self.val = 10 ** np.random.uniform(np.log10(self.p_1), np.log10(self.p_2))
        elif self.distribution == "normal":
            self.val = np.random.normal(self.p_1, self.p_2)
        elif self.distribution == "lognormal":
            self.val = 10 ** np.random.normal(self.p_1, self.p_2)
        else:
            raise RuntimeError(f"Unsupported distribution type: {self.distribution}.")
        self._ensure_step()

    def datatype(self):
        return int if self.step % 1 == 0 and self.p_1 % 1 == 0 else float


class ReferenceHyperparameter(Hyperparameter):
    """
    Hyperparameter that references only a different one.
    Thus, this parameter is not sampled but set after sampling.
    """

    def __init__(self, name: str, config: dict):
        super().__init__(name)
        self.reference = config.get("reference", None)

    def _ensure_step(self):
        """Make sure that the hyperparameter sticks to the discrete grid"""
        # nothing to do for references
        return

    def sample(self):
        """Sample this parameter, respecting properties"""
        # nothing to do for references
        return

    def datatype(self):
        raise RuntimeError("Datatype unknown for ReferenceHyperparameter")


class CategoricalHyperparameter(Hyperparameter):
    """
    A sampled hyperparameter, which is constraint to fixed,
    user given values and datatype
    """

    def __init__(self, name: str, config: dict):
        super().__init__(name)
        self.allowed_values = config["values"]
        if "datatype" not in config:
            raise RuntimeError(
                "Please specifiy datatype for all categorical hyper-parameters!, e.g. datatype=str"
            )
        self.type = locate(config["datatype"])
        self.allowed_values = [self.type(v) for v in self.allowed_values]

    def _ensure_step(self):
        """Make sure that the hyperparameter sticks to the discrete grid"""
        # nothing to do for categorical ones
        return

    def sample(self):
        """Sample this parameter, respecting properties"""
        # nothing to do for references
        idx = np.random.randint(0, len(self.allowed_values))
        self.val = self.allowed_values[idx]

    def datatype(self):
        return self.type


def get_hyperparameter(name: str, config: dict) -> Hyperparameter:
    """Factory function. Instantiates the correct Hyperparameter"""
    if "reference" in config.keys():
        return ReferenceHyperparameter(name, config)
    dist = config.get("distribution", None)
    if dist == "categorical":
        return CategoricalHyperparameter(name, config)

    return SampledHyperparameter(name, config)


def check_constraints(params: List[Hyperparameter], constraints) -> bool:
    """Check if the constraints are fulfilled."""
    # set each param as a local variable
    for par in params:
        locals().update({par.name: par.val})

    # set references
    for par in params:
        if isinstance(par, ReferenceHyperparameter):
            try:
                setattr(par, "val", eval(par.reference))
                # NOTE: literal_eval will cause ValueError: malformed node or string
            except Exception as ex:
                logger = Logger.get_logger()
                logger.info(f"error in evaluating expression: {par.reference}")
                raise ex
            locals().update({par.name: par.val})

    if constraints is None:
        return True  # shortcut

    # check all constraints
    for constr in constraints:
        try:
            const_res = eval(constr)
            # NOTE: literal_eval will cause ValueError: malformed node or string
        except SyntaxError as ex:
            raise SyntaxError(f"Invalid syntax in yaml config: {constr}") from ex
        if not const_res:
            return False

    return True


def sample_parameters(
    init_params: List[Hyperparameter],
    constraints,
    shared_config=None,
    shared_samples=None,
) -> dict:
    """
    Tries to sample from the hyperparameter list.

    Errors if in 10_0000 attempts no sample complying with the
    constraints is found.
    """
    for _ in range(10_000):
        params = copy.deepcopy(init_params)
        for par in params:
            par.sample()
        # add a random hyperparameter from the shared hyperparameter dataframe
        if shared_samples is not None:
            # sample one line from the pandas dataframe
            shared_samp = shared_samples.sample(1).iloc[0]["params"]
            for key in shared_samp.keys():
                par = Hyperparameter(key)
                par.val = shared_samp[key]
                par.name = key
                params.append(par)
        # check constrained
        if check_constraints(params, constraints):
            samples = {}
            for par in params:
                samples[par.name] = par.val
            return samples

    # if there was no sample found fullfilling the constrained above,
    # this may be due to the shared hyperparameters.
    # If so, new samples are generated for the shared hyperparameters
    logger = Logger.get_logger()
    logger.warning(
        "The constrainds coundn't be met with the shared Hyperparameters, "
        "shared dataframe pool will be ignored for now."
    )
    for _ in range(10_000):
        params = copy.deepcopy(init_params)
        # add the shared hyperparameter as a sampled hyperparameter
        if shared_samples is not None:
            shared_samp = shared_samples.sample(1).iloc[0]["params"]
            for key in shared_samp.keys():
                par = SampledHyperparameter(key, shared_config[key])
                par.sample()
                par.name = key
                params.append(par)
        for par in params:
            par.sample()
        # check constrained
        if check_constraints(params, constraints):
            samples = {}
            for par in params:
                samples[par.name] = par.val
            return samples

    raise RuntimeError(
        "Could not find an acceptable sample in 10,000 runs."
        "Are the bounds and constraints reasonable?"
    )


def create_samples_from_shared_samples(
    shared_samples: pd.DataFrame, config: dict, task_name: str
):
    """
    add informations like task, G_MODEL_NA and constrainds to the shared samples
    Parameters:
    shared_samples: pd Dataframe with columns [G_METHOD_NA, G_MODEL_NA, 'params']
    config: dataframe with yaml configuration of the current task
    task_name: name of the current task
    """
    shared_samp = shared_samples.copy()
    shared_samp[G_MODEL_NA] = config["model"]
    shared_samp[G_METHOD_NA] = task_name
    # respect the constraints if specified in the task
    if "constraints" in config.keys():
        for idx in range(shared_samp.shape[0] - 1, -1, -1):
            name = list(shared_samp["params"].iloc[idx].keys())[0]
            value = shared_samp["params"].iloc[idx][name]
            par = Hyperparameter(name)
            par.val = value
            if not check_constraints([par], config["constraints"]):
                shared_samp = shared_samp.drop(idx)
    return shared_samp


def sample_task_only_shared(
    num_samples, task_name, sample_df, config, shared_conf_samp
):
    """
    sample one task and add it to the dataframe for task descriptions which only
    contain shared hyperparameters
    """
    shared_config, shared_samples = shared_conf_samp
    # copy the shared samples dataframe and add the corrct G_MODEL_NA and taks names
    shared_samp = create_samples_from_shared_samples(shared_samples, config, task_name)

    # for the case that we expect more hyperparameter samples for the G_MODEL_NArithm as provided
    # in the shared sampes we use the shared config to sample new hyperparameters to ensure
    # that we have distinct hyperparameters
    if num_samples - shared_samp.shape[0] > 0:
        s_config = shared_config.copy()
        s_dict = {}
        for keys in s_config.keys():
            if keys != "num_shared_param_samples":
                s_dict[keys] = s_config[keys]
        if "constraints" in config.keys():
            s_dict["constraints"] = config["constraints"]
        s_config["model"] = config["model"]
        s_config["hyperparameters"] = s_dict

        # sample new shared hyperparameters
        sample_df = sample_task(
            num_samples - shared_samp.shape[0],
            task_name,
            (s_config, sample_df),
            (None, None),
        )
        # add previously sampled shared hyperparameters
        sample_df = sample_df.append(shared_samp, ignore_index=True)
    # for the case that the number of shared samples is >= the expected number of
    # sampled hyperparameters we randomly choose rows in the sampled hyperparameters df
    else:
        shared_samp = shared_samp.sample(num_samples)
        sample_df = sample_df.append(shared_samp, ignore_index=True)

    return sample_df


def sample_task(
    num_samples: int, task_name: str, conf_samp: tuple, shared_conf_samp: tuple
):
    """Sample one task and add it to the dataframe"""
    config, sample_df = conf_samp
    shared_config, shared_samples = shared_conf_samp
    if "hyperparameters" in config.keys():
        # in benchmark configuration file, sub-section hyperparameters
        # means changing hyper-parameters
        params = []
        for key, val in config["hyperparameters"].items():
            if key in ("constraints", "num_shared_param_samples"):
                continue
            params += [get_hyperparameter(key, val)]

        constraints = config["hyperparameters"].get("constraints", None)
        for _ in range(num_samples):
            sample = sample_parameters(
                params, constraints, shared_config, shared_samples
            )
            sample_df.loc[len(sample_df.index)] = [task_name, config["model"], sample]
    elif "shared" in config.keys():
        sample_df = sample_task_only_shared(
            num_samples, task_name, sample_df, config, (shared_config, shared_samples)
        )
    else:
        # add single line if no varying hyperparameters are specified.
        sample_df.loc[len(sample_df.index)] = [task_name, config["model"], {}]
    return sample_df


def is_dict_with_key(input_dict, key) -> bool:
    """Determines if the input argument is a dictionary and it has key"""
    return isinstance(input_dict, dict) and key in input_dict.keys()


def get_shared_samples(
    shared_samples_full: pd.DataFrame, shared_config_full: dict, task_config: dict
):
    """
    - creates a dataframe with columns [task, G_MODEL_NA, params],
    task and G_MODEL_NA are all for all rows, but params is filled with the
    shared parameters of shared_samples_full requested by task_config.

    - creates a shared config containing only information about the
    shared hyperparameters requested by the task_config
    """
    shared_samples = shared_samples_full.copy(deep=True)
    shared_config = shared_config_full.copy()
    if "shared" in task_config.keys():
        shared = task_config["shared"]
    else:
        shared = []
    for line_num in range(shared_samples.shape[0]):
        hyper_p_dict = shared_samples.iloc[line_num]["params"].copy()
        key_list = copy.deepcopy(list(hyper_p_dict.keys()))
        for key_ in key_list:
            if key_ not in shared:
                del hyper_p_dict[key_]
        shared_samples.iloc[line_num]["params"] = hyper_p_dict
    for key_ in key_list:
        if not key_ == "num_shared_param_samples":
            if key_ not in shared:
                del shared_config[key_]
    # remove all duplicates
    shared_samples = shared_samples.drop_duplicates(subset="params")
    return shared_samples, shared_config


def sample_hyperparameters(
    config: dict, dest: str = None, sampling_seed: int = None
) -> pd.DataFrame:
    """
    Samples the hyperparameters according to the given
    config, which should be the dictionary of the full
    benchmark config yaml.
    Result is saved to 'output_dir/hyperparameters.csv' of the
    config if not specified explicitly.

    Note: Parts of the yaml content are executed. Thus use this
    only with trusted config files.
    """
    if dest is None:
        dest = config["output_dir"] + os.sep + "hyperparameters.csv"

    if sampling_seed is not None:
        np.random.seed(sampling_seed)

    num_samples = config["num_param_samples"]
    samples = pd.DataFrame(columns=[G_METHOD_NA, G_MODEL_NA, "params"])
    if "Shared params" in config.keys():
        shared_config_full = config["Shared params"]
        shared_samples_full = pd.DataFrame(columns=[G_METHOD_NA, G_MODEL_NA, "params"])
        shared_val = {"model": "all", "hyperparameters": config["Shared params"]}
        # fill up the dataframe shared samples
        shared_samples_full = sample_task(
            shared_config_full["num_shared_param_samples"],
            "all",
            (shared_val, shared_samples_full),
            (None, None),
        )
    else:
        shared_samples_full = None
    for key, val in config.items():
        if is_dict_with_key(val, "model"):
            if shared_samples_full is not None:
                shared_samples, shared_config = get_shared_samples(
                    shared_samples_full, shared_config_full, val
                )
            else:
                shared_config = None
                shared_samples = None

            samples = sample_task(
                num_samples, key, (val, samples), (shared_config, shared_samples)
            )

    os.makedirs(os.path.dirname(dest), exist_ok=True)

    # create a txt file with the commit information
    with open(
        config["output_dir"] + os.sep + "commit.txt", "w", encoding="utf8"
    ) as file:
        file.writelines("use git log |grep \n")
        file.writelines("consider remove leading b in the line below \n")
        file.write(get_git_tag())
    with open(
        config["output_dir"] + os.sep + "config.txt", "w", encoding="utf8"
    ) as file:
        json.dump(config, file)

    samples.to_csv(dest)
    return samples



"""
construct feature extractor, task neural network (e.g. classification) and domain classification
network
"""
from torch.nn import functional as F

from domainlab import g_str_cross_entropy_agg
from domainlab.compos.nn_zoo.net_adversarial import AutoGradFunReverseMultiply
from domainlab.models.a_model_classif import AModelClassif


def mk_dann(parent_class=AModelClassif):
    """
    Instantiate a Deep Adversarial Net (DAN) model

    Details:
        The model is trained to solve two tasks:
        1. Standard image classification.
        2. Domain classification.
        Here for, a feature extractor is adversarially trained to minimize the loss of the image
        classifier and maximize the loss of the domain classifier.
        For more details, see:
        Ganin, Yaroslav, et al. "Domain-adversarial training of neural networks."
        The journal of machine learning research 17.1 (2016): 2096-2030.

    Args:
        parent_class (AModel, optional): Class object determining the task
        type. Defaults to AModelClassif.

    Returns:
        ModelDAN: model inheriting from parent class

    Input Parameters:
        list_str_y: list of labels,
        list_d_tr: list of training domains
        alpha: total_loss = task_loss + $$\\alpha$$ * domain_classification_loss,
        net_encoder: neural network to extract the features (input: training data),
        net_classifier: neural network (input: output of net_encoder; output: label prediction),
        net_discriminator: neural network (input: output of net_encoder;
        output: prediction of training domain)

    Usage:
        For a concrete example, see:
        https://github.com/marrlab/DomainLab/blob/master/tests/test_mk_exp_dann.py
    """

    class ModelDAN(parent_class):
        """
        anonymous
        """

        def __init__(
            self,
            list_str_y,
            list_d_tr,
            alpha,
            net_encoder,
            net_classifier,
            net_discriminator,
            builder=None,
        ):
            """
            See documentation above in mk_dann() function
            """
            super().__init__(list_str_y)
            self.list_d_tr = list_d_tr
            self.alpha = alpha
            self._net_invar_feat = net_encoder
            self._net_classifier = net_classifier
            self.net_discriminator = net_discriminator
            self.builder = builder

        def reset_aux_net(self):
            """
            reset auxilliary neural network: domain classifier
            """
            if self.builder is None:
                return
            self.net_discriminator = self.builder.reset_aux_net(
                self.extract_semantic_feat
            )

        def hyper_update(self, epoch, fun_scheduler):
            """hyper_update.
            :param epoch:
            :param fun_scheduler: the hyperparameter scheduler object
            """
            dict_rst = fun_scheduler(
                epoch
            )  # the __call__ method of hyperparameter scheduler
            self.alpha = dict_rst["alpha"]

        def hyper_init(self, functor_scheduler):
            """hyper_init.
            :param functor_scheduler:
            """
            return functor_scheduler(trainer=None, alpha=self.alpha)

        def _cal_reg_loss(self, tensor_x, tensor_y, tensor_d, others):
            _ = others
            _ = tensor_y
            feat = self.extract_semantic_feat(tensor_x)
            net_grad_additive_reverse = AutoGradFunReverseMultiply.apply(
                feat, self.alpha
            )
            logit_d = self.net_discriminator(net_grad_additive_reverse)
            _, d_target = tensor_d.max(dim=1)
            lc_d = F.cross_entropy(logit_d, d_target, reduction=g_str_cross_entropy_agg)
            return [lc_d], [self.alpha]

    return ModelDAN



import subprocess
import warnings
from subprocess import CalledProcessError

from domainlab.utils.logger import Logger


def get_git_tag(print_diff=False):
    flag_not_commited = False
    logger = Logger.get_logger()
    try:
        subprocess.check_output(["git", "diff-index", "--quiet", "HEAD"])
    except CalledProcessError:
        logger.warning("\n\n")
        logger.warning("!!!: not committed yet")
        warnings.warn("!!!: not committed yet")
        flag_not_commited = True
        logger.warning("\n\n")
        try:
            diff_byte = subprocess.check_output(["git", "diff"])
            if print_diff:
                logger.info(str(diff_byte))  # print is currently ugly, do not use!
        except Exception:
            logger = Logger.get_logger()
            logger.warning("not in a git repository")
            warnings.warn("not in a git repository")
    try:
        tag_byte = subprocess.check_output(["git", "describe", "--always"]).strip()
        logger.info(str(tag_byte))
        tag_str = str(tag_byte)
        git_str = tag_str.replace("'", "")
        if flag_not_commited:
            git_str = git_str + "_not_commited"
        return git_str
    except Exception:
        logger = Logger.get_logger()
        logger.warning("not in a git repository")
        warnings.warn("not in a git repository")
    return "no_git_version"



"""
Model Selection should be decoupled from
"""
import math

from domainlab.algos.msels.a_model_sel import AMSel
from domainlab.utils.logger import Logger


class MSelTrLoss(AMSel):
    """
    1. Model selection using sum of loss across training domains
    2. Visitor pattern to trainer
    """

    def __init__(self, max_es):
        super().__init__()
        # NOTE: super() must come first otherwise it will overwrite existing
        # values!
        self.reset()
        self._max_es = max_es

    def reset(self):
        self.best_loss = float("inf")
        self.es_c = 0

    @property
    def max_es(self):
        return self._max_es

    def update(self, clear_counter=False):
        """
        if the best model should be updated
        """
        loss = self.trainer.epo_loss_tr  # @FIXME
        assert loss is not None
        assert not math.isnan(loss)
        flag = True
        if loss < self.best_loss:
            self.es_c = 0  # restore counter
            self.best_loss = loss
        else:
            self.es_c += 1
            logger = Logger.get_logger()
            logger.info(f"early stop counter: {self.es_c}")
            logger.info(f"loss:{loss}, best loss: {self.best_loss}")
            flag = False  # do not update best model
            if clear_counter:
                logger.info("clearing counter")
                self.es_c = 0
        return flag

    def if_stop(self):
        """
        if should early stop
        """
        return self.es_c >= self.max_es



import torch
import torch.distributions as dist
from torch import nn

from domainlab.compos.utils_conv_get_flat_dim import get_flat_dim
from domainlab.compos.zoo_nn import FeatExtractNNBuilderChainNodeGetter


class EncoderConnectLastFeatLayer2Z(nn.Module):
    """
    Connect the last layer of a feature extraction
    neural network to the latent representation
    This class should be transparent to where to fetch the network
    """

    def __init__(
        self, z_dim, flag_pretrain, i_c, i_h, i_w, args, arg_name, arg_path_name
    ):
        """__init__.
        :param hidden_size:
        """
        super().__init__()
        net_builder = FeatExtractNNBuilderChainNodeGetter(
            args, arg_name, arg_path_name
        )()  # request

        self.net_feat_extract = net_builder.init_business(
            flag_pretrain=flag_pretrain,
            dim_out=z_dim,
            remove_last_layer=True,
            args=args,
            isize=(i_c, i_h, i_w),
        )

        size_last_layer_before_z = get_flat_dim(self.net_feat_extract, i_c, i_h, i_w)

        self.net_fc_mean = nn.Sequential(nn.Linear(size_last_layer_before_z, z_dim))
        self.net_fc_scale = nn.Sequential(
            nn.Linear(size_last_layer_before_z, z_dim), nn.Softplus()
        )  # for scale calculation

        torch.nn.init.xavier_uniform_(self.net_fc_mean[0].weight)
        self.net_fc_mean[0].bias.data.zero_()
        torch.nn.init.xavier_uniform_(self.net_fc_scale[0].weight)
        self.net_fc_scale[0].bias.data.zero_()

    def forward(self, x):
        """
        :param x:
        """
        feature = self.net_feat_extract(x)
        zd_q_loc = self.net_fc_mean(feature)
        zd_q_scale = self.net_fc_scale(feature) + 1e-7
        q_zd = dist.Normal(zd_q_loc, zd_q_scale)
        zd_q = q_zd.rsample()  # Reparameterization trick
        return q_zd, zd_q



"""
Base Class for XYD VAE
"""
import torch
import torch.distributions as dist

from domainlab.utils.utils_class import store_args


class InterfaceVAEXYD:
    """
    Interface (without constructor and inheritance) for XYD VAE
    """

    def init(self):
        self.chain_node_builder.init_business(self.zd_dim, self.zx_dim, self.zy_dim)
        self.i_c = self.chain_node_builder.i_c
        self.i_h = self.chain_node_builder.i_h
        self.i_w = self.chain_node_builder.i_w
        self._init_components()

    def _init_components(self):
        """
        q(z|x)
        p(zy)
        q_{classif}(zy)
        """
        self.add_module("encoder", self.chain_node_builder.build_encoder())
        self.add_module("decoder", self.chain_node_builder.build_decoder())
        self.add_module(
            "net_p_zy",
            self.chain_node_builder.construct_cond_prior(self.dim_y, self.zy_dim),
        )

    def init_p_zx4batch(self, batch_size, device):
        """
        1. Generate pytorch distribution object.
        2. To be called by trainer

        :param batch_size:
        :param device:
        """
        # p(zx): isotropic gaussian
        zx_p_loc = torch.zeros(batch_size, self.zx_dim).to(device)
        zx_p_scale = torch.ones(batch_size, self.zx_dim).to(device)
        p_zx = dist.Normal(zx_p_loc, zx_p_scale)
        return p_zx



"""
Runs one task for a single hyperparameter sample for each leave-out-domain
and each random seed.
"""
import ast
import copy
import gc

import numpy as np
import pandas as pd
import torch

from domainlab.arg_parser import apply_dict_to_args, mk_parser_main
from domainlab.exp.exp_cuda_seed import set_seed
from domainlab.exp.exp_main import Exp
from domainlab.exp.exp_utils import ExpProtocolAggWriter
from domainlab.utils.hyperparameter_sampling import G_METHOD_NA
from domainlab.utils.logger import Logger


def load_parameters(file: str, index: int) -> tuple:
    """
    Loads a single parameter sample
    @param file: csv file
    @param index: index of hyper-parameter
    """
    param_df = pd.read_csv(file, index_col=0)
    row = param_df.loc[index]
    params = ast.literal_eval(row.params)
    # row.task has nothing to do with DomainLab task, it is
    # benchmark task which correspond to one algorithm
    return row[G_METHOD_NA], params


def convert_dict2float(dict_in):
    """
    convert scientific notation from 1e5 to 10000
    """
    dict_out = copy.deepcopy(dict_in)
    for key, val in dict_out.items():
        if isinstance(val, str):
            try:
                val_float = float(val)
                dict_out[key] = val_float
            except:
                pass
    return dict_out


def run_experiment(
    config: dict,
    param_file: str,
    param_index: int,
    out_file: str,
    start_seed=None,
    misc=None,
    num_gpus=1,
):
    """
    Runs the experiment several times:

    for test_domain in test_domains:
        for seed from startseed to endseed:
            evaluate the algorithm with test_domain, initialization with seed

    :param config: dictionary from the benchmark yaml
    :param param_file: path to the csv with the parameter samples
    :param param_index: parameter index that should be covered by this task,
    currently this correspond to the line number in the csv file, or row number
    in the resulting pandas dataframe
    :param out_file: path to the output csv
    :param start_seed: random seed to start for stochastic variations of pytorch
    :param misc: optional dictionary of additional parameters, if any.

    # FIXME: we might want to run the experiment using commandline arguments
    """

    if misc is None:
        misc = {}
    str_algo_as_task, hyperparameters = load_parameters(param_file, param_index)
    logger = Logger.get_logger()
    logger.debug(
        "\n*******************************************************************"
    )
    logger.debug(
        f"{str_algo_as_task}, param_index={param_index}, params={hyperparameters}"
    )
    logger.debug(
        "*******************************************************************\n"
    )
    misc["result_file"] = out_file
    misc["params"] = hyperparameters
    misc["benchmark_task_name"] = str_algo_as_task
    misc["param_index"] = param_index
    misc["keep_model"] = False

    parser = mk_parser_main()
    args = parser.parse_args(args=[])
    args_algo_specific = config[str_algo_as_task].copy()
    if "hyperparameters" in args_algo_specific:
        del args_algo_specific["hyperparameters"]
    args_domainlab_common_raw = config.get("domainlab_args", {})
    args_domainlab_common = convert_dict2float(args_domainlab_common_raw)
    # check if some of the hyperparameters are already specified
    # in args_domainlab_common or args_algo_specific
    if (
        np.intersect1d(
            list(args_algo_specific.keys()), list(hyperparameters.keys())
        ).shape[0]
        > 0
    ):
        logger.error(
            f"the hyperparameter "
            f"{np.intersect1d(list(args_algo_specific.keys()), list(hyperparameters.keys()))}"
            f" has already been fixed to a value in the algorithm section."
        )
        raise RuntimeError(
            f"the hyperparameter "
            f"{np.intersect1d(list(args_algo_specific.keys()), list(hyperparameters.keys()))}"
            f" has already been fixed to a value in the algorithm section."
        )
    if (
        np.intersect1d(
            list(args_domainlab_common.keys()), list(hyperparameters.keys())
        ).shape[0]
        > 0
    ):
        logger.error(
            f"the hyperparameter "
            f"{np.intersect1d(list(args_algo_specific.keys()), list(hyperparameters.keys()))}"
            f" has already been fixed to a value in the domainlab_args section."
        )
        raise RuntimeError(
            f"the hyperparameter "
            f"{np.intersect1d(list(args_algo_specific.keys()), list(hyperparameters.keys()))}"
            f" has already been fixed to a value in the domainlab_args section."
        )
    apply_dict_to_args(args, args_domainlab_common)
    args_algo_specific_scientific_notation = convert_dict2float(args_algo_specific)
    apply_dict_to_args(args, args_algo_specific_scientific_notation, extend=True)
    apply_dict_to_args(args, hyperparameters)
    apply_dict_to_args(args, misc, extend=True)
    gpu_ind = param_index % num_gpus
    args.device = str(gpu_ind)

    if torch.cuda.is_available():
        torch.cuda.init()
        logger.info("before experiment loop: ")
        logger.info(str(torch.cuda.memory_summary()))
    if start_seed is None:
        start_seed = config["startseed"]
        end_seed = config["endseed"]
    else:
        end_seed = start_seed + (config["endseed"] - config["startseed"])
    for seed in range(start_seed, end_seed + 1):
        for te_d in config["test_domains"]:
            args.te_d = te_d
            set_seed(seed)
            args.seed = seed
            try:
                if torch.cuda.is_available():
                    logger.info("before experiment starts")
                    logger.info(str(torch.cuda.memory_summary()))
            except KeyError as ex:
                logger.error(str(ex))
            args.lr = float(args.lr)
            # <=' not supported between instances of 'float' and 'str
            exp = Exp(args=args, visitor=ExpProtocolAggWriter)
            # NOTE: if key "testing" is set in benchmark, then do not execute
            # experiment
            if not misc.get("testing", False):
                exp.execute()
            try:
                if torch.cuda.is_available():
                    logger.info("before torch memory clean up")
                    logger.info(str(torch.cuda.memory_summary()))
            except KeyError as ex:
                logger.error(str(ex))
            del exp
            torch.cuda.empty_cache()
            gc.collect()
            try:
                if torch.cuda.is_available():
                    logger.info("after torch memory clean up")
                    logger.info(str(torch.cuda.memory_summary()))
            except KeyError as ex:
                logger.error(str(ex))



"""
Model Selection should be decoupled from
"""
from domainlab.algos.msels.a_model_sel import AMSel
from domainlab.utils.logger import Logger


class MSelOracleVisitor(AMSel):
    """
    save best out-of-domain test acc model, but do not affect
    how the final model is selected
    """

    def __init__(self, msel=None):
        """
        Decorator pattern
        """
        super().__init__()
        self.best_oracle_acc = 0
        self.msel = msel

    @property
    def oracle_last_setpoint_sel_te_acc(self):
        """
        last setpoint acc
        """
        if self.msel is not None and hasattr(
            self.msel, "oracle_last_setpoint_sel_te_acc"
        ):
            return self.msel.oracle_last_setpoint_sel_te_acc
        return -1

    def update(self, clear_counter=False):
        """
        if the best model should be updated
        """
        self.trainer.model.save("epoch")
        flag = False
        if self.tr_obs.metric_val is None:
            return super().update(clear_counter)
        metric = self.tr_obs.metric_te[self.tr_obs.str_metric4msel]
        if metric > self.best_oracle_acc:
            self.best_oracle_acc = metric
            if self.msel is not None:
                self.trainer.model.save("oracle")
            else:
                self.trainer.model.save()
            logger = Logger.get_logger()
            logger.info("new oracle model saved")
            flag = True
        if self.msel is not None:
            return self.msel.update(clear_counter)
        return flag

    def if_stop(self):
        """
        if should early stop
        oracle model selection does not intervene how models get selected
        by the innermost model selection
        """
        if self.msel is not None:
            return self.msel.if_stop()
        return False

    def accept(self, trainer, tr_obs):
        if self.msel is not None:
            self.msel.accept(trainer, tr_obs)
        super().accept(trainer, tr_obs)



"""
hyper-parameters for JiGen
"""


def add_args2parser_jigen(parser):
    """
    hyper-parameters for JiGen
    """
    parser.add_argument("--nperm", type=int, default=31, help="number of permutations")
    parser.add_argument(
        "--pperm",
        type=float,
        default=0.1,
        help="probability of permutating the tiles \
                        of an image",
    )
    parser.add_argument(
        "--jigen_ppath",
        type=str,
        default=None,
        help="npy file path to load numpy array with each row being \
                        permutation index, if not None, nperm and grid_len has to agree \
                        with the number of row and columns of the input array",
    )
    parser.add_argument(
        "--grid_len", type=int, default=3, help="length of image in tile unit"
    )
    return parser



"""
Integrate Chain-of-Responsibility and Builder Pattern for feature extract
"""

from domainlab.compos.pcr.p_chain_handler import AbstractChainNodeHandler
from domainlab.utils.utils_class import store_args


class AbstractFeatExtractNNBuilderChainNode(AbstractChainNodeHandler):
    """
    to ensure chain of responsibility node AbstractChainNodeHandler always
    work even some node can not start their heavy weight business object,
    avoid override the initializer so that node construction is always
    light weight.
    """

    def __init__(self, successor_node):
        """__init__.

        :param successor_node:
        """
        self.net_feat_extract = None
        super().__init__(successor_node)

    @store_args
    def init_business(
        self,
        dim_out,
        args,
        i_c=None,
        i_h=None,
        i_w=None,
        flag_pretrain=None,
        remove_last_layer=False,
    ):
        """
        initialize **and** return the heavy weight business object for doing
        the real job
        :param request: subclass can override request object to be string or
        function
        :return: the constructed service object
        """
        return NotImplementedError

    def is_myjob(self, args):
        """
        :param args_nname: command line arguments:
            "--nname": name of the torchvision model
            "--npath": path to the user specified python file with neural
            network definition
        """
        return NotImplementedError



from domainlab.utils.utils_class import store_args


class RequestVAEBuilderCHW:
    @store_args
    def __init__(self, i_c, i_h, i_w, args):
        pass


class RequestVAEBuilderNN:
    """creates request when input does not come from command-line (args) but from test_exp file"""

    @store_args
    def __init__(self, net_class_d, net_x, net_class_y, i_c, i_h, i_w):
        """net_class_d, net_x and net_class_y are neural networks defined by the user"""


class RequestTask:
    """
    Isolate args from Request object of chain of responsibility node for task
    """

    def __init__(self, args):
        self.args = args

    def __call__(self):
        return self.args.task


class RequestArgs2ExpCmd:
    """
    Isolate args from Request object of chain of responsibility node for experiment
    For example, args has field names which will couple with experiment class, this
    request class also serves as isolation class or adaptation class
    """

    @store_args
    def __init__(self, args):
        self.args = args

    def __call__(self):
        return self.args.model



import abc

import torch

from domainlab.models.a_model_classif import AModelClassif
from domainlab.utils.utils_classif import get_label_na, logit2preds_vpic


class AModelCustom(AModelClassif):
    """AModelCustom."""

    def __init__(self, list_str_y, list_str_d=None):
        """__init__.
        :param net:
        :param list_str_y:
        :param list_str_d:
        """
        super().__init__(list_str_y)

    @property
    @abc.abstractmethod
    def dict_net_module_na2arg_na(self):
        """dict_net_module_na2arg_na.
        A dictionary with the key being the pytorch module name and value
        being the commandline argument name
        """
        raise NotImplementedError

    def extract_semantic_feat(self, tensor_x):
        """
        calculate the logit for softmax classification
        """
        raise NotImplementedError



"""
Emperical risk minimization
"""
from domainlab.compos.nn_zoo.nn import LayerId
from domainlab.models.a_model_classif import AModelClassif
from domainlab.utils.override_interface import override_interface


def mk_erm(parent_class=AModelClassif):
    """
    Instantiate a Deepall (ERM) model

    Details:
        Creates a model, which trains a neural network via standard empirical risk minimization
        (ERM). The fact that the training data stems from different domains is neglected, as all
        domains are pooled together during training.

    Args:
        parent_class (AModel, optional):
            Class object determining the task type. Defaults to AModelClassif.

    Returns:
        ModelERM: model inheriting from parent class

    Input Parameters:
        custom neural network, the output dimension must be the number of labels

    Usage:
        For a concrete example, see:
        https://github.com/marrlab/DomainLab/blob/tests/test_mk_exp_erm.py
    """

    class ModelERM(parent_class):
        """
        anonymous
        """

        def __init__(
            self, net=None, net_feat=None, net_classifier=None, list_str_y=None
        ):
            if net_feat is None and net_classifier is None and net is not None:
                net_feat = net
                net_classifier = LayerId()
                dim_y = list(net.modules())[-1].out_features
            elif net_classifier is not None:
                dim_y = list(net_classifier.modules())[-1].out_features
            else:
                raise RuntimeError(
                    "specify either a whole network for classification or separate \
                        feature and classifier"
                )
            if list_str_y is None:
                list_str_y = [f"class{i}" for i in range(dim_y)]
            super().__init__(list_str_y)
            self._net_classifier = net_classifier
            self._net_invar_feat = net_feat

    return ModelERM



"""
Bridge Pattern: Separation of interface and implementation.
This class is using one implementation to feed into parent class constructor.
"""
import numpy as np
import torch.nn as nn

from domainlab.compos.nn_zoo.net_gated import Conv2d, GatedConv2d, GatedDense
from domainlab.compos.vae.compos.decoder_concat_vec_reshape_conv import (
    DecoderConcatLatentFcReshapeConv,
)
from domainlab.compos.vae.compos.decoder_losses import NLLPixelLogistic256


class DecoderConcatLatentFCReshapeConvGatedConv(DecoderConcatLatentFcReshapeConv):
    """
    Bridge Pattern: Separation of interface and implementation.
    This class is using implementation to feed into parent class constructor.
    Latent vector re-arranged to image-size directly, then convolute
    to get the textures of the original image
    """

    def __init__(self, z_dim, i_c, i_h, i_w):
        """
        :param z_dim:
        :param list_im_chw: [im_c, im_h, im_w]
        im_c, im_h, im_w: [channel, height, width]
        """
        list_im_chw = [i_c, i_h, i_w]
        cls_fun_nll_p_x = NLLPixelLogistic256
        net_fc_z2flat_img = nn.Sequential(GatedDense(z_dim, np.prod(list_im_chw)))

        net_conv = nn.Sequential(
            # GatedConv2d
            # input_channels, output_channels, kernel_size,
            # stride, padding, dilation=1, activation=None)
            GatedConv2d(list_im_chw[0], 64, 3, 1, 1),
            GatedConv2d(64, 64, 3, 1, 1),
            # GatedConv2d(64, 64, 3, 1, 1),
            # comment out 2 layers to reduce decoder power
            # GatedConv2d(64, 64, 3, 1, 1),
        )
        #
        # hidden image to mean and variance of each pixel
        # stride(1) and kernel size 1, pad 0
        net_p_x_mean = Conv2d(64, list_im_chw[0], 1, 1, 0, activation=nn.Sigmoid())
        net_p_x_log_var = Conv2d(
            64,
            list_im_chw[0],
            1,
            1,
            0,
            activation=nn.Hardtanh(min_val=-4.5, max_val=0.0),
        )
        super().__init__(
            z_dim,
            i_c,
            i_h,
            i_w,
            cls_fun_nll_p_x,
            net_fc_z2flat_img,
            net_conv,
            net_p_x_mean,
            net_p_x_log_var,
        )



"""
create dictionary for matching
"""
import torch

from domainlab.utils.logger import Logger


class MatchDictInit:
    """
    base class for matching dictionary creator
    """

    def __init__(self, keys, vals, i_c, i_h, i_w):
        self.keys = keys
        self.vals = vals
        self.i_c = i_c
        self.i_h = i_h
        self.i_w = i_w

    def get_num_rows(self, key):
        raise NotImplementedError

    def __call__(self):
        dict_data = {}
        for key in self.keys:
            dict_data[key] = {}
            num_rows = self.get_num_rows(key)
            dict_data[key]["data"] = torch.rand(
                (num_rows, self.i_c, self.i_w, self.i_h)
            )
            # @FIXME: some labels won't be filled at all, when using training loader since the incomplete batch is dropped
            dict_data[key]["label"] = torch.rand((num_rows, 1))  # scalar label
            dict_data[key]["idx"] = torch.randint(low=0, high=1, size=(num_rows, 1))
        return dict_data


class MatchDictVirtualRefDset2EachDomain(MatchDictInit):
    """
    dict[0:virtual_ref_dset_size] has tensor dimension: (num_domains_tr, i_c, i_h, i_w)
    """

    def __init__(self, virtual_ref_dset_size, num_domains_tr, i_c, i_h, i_w):
        """
        virtual_ref_dset_size is a virtual dataset, len(virtual_ref_dset_size) = sum of all popular domains
        """
        super().__init__(
            keys=range(virtual_ref_dset_size),
            vals=num_domains_tr,
            i_c=i_c,
            i_h=i_h,
            i_w=i_w,
        )

    def get_num_rows(self, key=None):
        """
        key is 0:virtual_ref_dset_size
        """
        return self.vals  # total_domains


class MatchDictNumDomain2SizeDomain(MatchDictInit):
    """
    tensor dimension for the kth domain: [num_domains_tr, (size_domain_k, i_c, i_h, i_w)]
    """

    def __init__(self, num_domains_tr, list_tr_domain_size, i_c, i_h, i_w):
        super().__init__(
            keys=range(num_domains_tr),
            vals=list_tr_domain_size,
            i_c=i_c,
            i_h=i_h,
            i_w=i_w,
        )

    def get_num_rows(self, key):
        return self.vals[key]  # list_tr_domain_size[domain_index]


def dist_cosine_agg(x1, x2):
    """
    torch.nn.CosineSimilarity assumes x1 and x2 share exactly the same dimension
    """
    fun_cos = torch.nn.CosineSimilarity(dim=1, eps=1e-08)
    return 1.0 - fun_cos(x1, x2)


def fun_tensor_normalize(tensor_batch_x):
    eps = 1e-8
    batch_norm_x = tensor_batch_x.norm(
        dim=1
    )  # Frobenius norm or Euclidean Norm long the embedding direction, len(norm) should be batch_size
    batch_norm_x = batch_norm_x.view(
        batch_norm_x.shape[0], 1
    )  # add dimension to tensor
    tensor_eps = eps * torch.ones_like(batch_norm_x)
    tensor_batch_x = tensor_batch_x / torch.max(batch_norm_x, tensor_eps)
    assert not torch.sum(torch.isnan(tensor_batch_x))
    return tensor_batch_x


def dist_pairwise_cosine(x1, x2, tau=0.05):
    """
    x1 and x2 does not necesarilly have the same shape, and we want to have a cartesian product of the pairwise distances
    """
    assert len(x1.shape) == 2 and len(x2.shape) == 2
    assert not torch.sum(torch.isnan(x1))
    assert not torch.sum(torch.isnan(x2))

    x1 = fun_tensor_normalize(x1)
    x2 = fun_tensor_normalize(x2)

    x1_extended_dim = x1.unsqueeze(
        1
    )  # Returns a new tensor with a dimension of size one inserted at the specified position.
    # extend the order of by insering a new dimension so that cartesion product of pairwise distance can be calculated

    # since the batch size of x1 and x2 won't be the same, directly calculting elementwise product will cause an error
    # with order 3 multiply order 2 tensor, the feature dimension will be matched then the rest dimensions form cartesian product
    cos_sim = torch.sum(x1_extended_dim * x2, dim=2)  # elementwise product
    cos_sim = cos_sim / tau  # make cosine similarity bigger than 1
    assert not torch.sum(torch.isnan(cos_sim))
    loss = torch.sum(torch.exp(cos_sim), dim=1)
    assert not torch.sum(torch.isnan(loss))
    return loss


def get_base_domain_size4match_dg(task):
    """
    Base domain is a dataset where each class
    set come from one of the nominal domains
    """
    # @FIXME: base domain should be calculated only on training domains
    # instead of all the domains!
    # domain_keys = task.get_list_domains()
    domain_keys = task.list_domain_tr
    base_domain_size = 0
    classes = task.list_str_y
    for mclass in classes:
        num = 0
        ref_domain = -1
        for _, domain_key in enumerate(domain_keys):
            if task.dict_domain_class_count[domain_key][mclass] > num:
                ref_domain = domain_key
                num = task.dict_domain_class_count[domain_key][mclass]
        logger = Logger.get_logger()
        logger.info(
            f"for class {mclass} bigest sample size is {num} "
            f"ref domain is {ref_domain}"
        )
        base_domain_size += num
    return base_domain_size



"""
trainer matchdg
"""
import torch

from domainlab import g_inst_component_loss_agg, g_list_loss_agg
from domainlab.algos.trainers.a_trainer import AbstractTrainer
from domainlab.algos.trainers.compos.matchdg_match import MatchPair
from domainlab.algos.trainers.compos.matchdg_utils import (
    dist_cosine_agg,
    dist_pairwise_cosine,
    get_base_domain_size4match_dg,
)
from domainlab.tasks.utils_task_dset import DsetIndDecorator4XYD
from domainlab.utils.logger import Logger


class TrainerMatchDG(AbstractTrainer):
    """
    Contrastive Learning
    """

    def dset_decoration_args_algo(self, args, ddset):
        ddset = DsetIndDecorator4XYD(ddset)
        return ddset

    def init_business(
        self, model, task, observer, device, aconf, flag_accept=True, flag_erm=False
    ):
        """
        initialize member objects
        """
        super().init_business(model, task, observer, device, aconf, flag_accept)
        # use the same batch size for match tensor
        # so that order is kept!
        self.base_domain_size = get_base_domain_size4match_dg(self.task)
        self.epo_loss_tr = 0
        self.flag_erm = flag_erm
        self.lambda_ctr = self.aconf.gamma_reg
        self.mk_match_tensor(epoch=0)
        self.flag_match_tensor_sweep_over = False
        self.tuple_tensor_ref_domain2each_y = None
        self.tuple_tensor_refdomain2each = None

    def tr_epoch(self, epoch):
        """
        # data in one batch comes from two sources: one part from loader,
        # the other part from match tensor
        """
        self.model.train()
        self.epo_loss_tr = 0
        logger = Logger.get_logger()
        # update match tensor
        if (epoch + 1) % self.aconf.epos_per_match_update == 0:
            self.mk_match_tensor(epoch)

        inds_shuffle = torch.randperm(self.tensor_ref_domain2each_domain_x.size(0))
        # NOTE: match tensor size: N(ref domain size) * #(train domains) * (image size: c*h*w)
        # self.tensor_ref_domain2each_domain_x[inds_shuffle]
        # shuffles the match tensor at the first dimension
        self.tuple_tensor_refdomain2each = torch.split(
            self.tensor_ref_domain2each_domain_x[inds_shuffle], self.aconf.bs, dim=0
        )
        # Splits the tensor into chunks.
        # Each chunk is a view of the original tensor of batch size self.aconf.bs
        # return is a tuple of the splited chunks
        self.tuple_tensor_ref_domain2each_y = torch.split(
            self.tensor_ref_domain2each_domain_y[inds_shuffle], self.aconf.bs, dim=0
        )
        logger.info(
            f"number of batches in match tensor: {len(self.tuple_tensor_refdomain2each)}"
        )
        logger.info(
            f"single batch match tensor size: {self.tuple_tensor_refdomain2each[0].shape}"
        )

        for batch_idx, (x_e, y_e, d_e, *others) in enumerate(self.loader_tr):
            # random loader with same batch size as the match tensor loader
            # the 4th output of self.loader is not used at all,
            # is only used for creating the match tensor
            self.tr_batch(epoch, batch_idx, x_e, y_e, d_e, others)
            if self.flag_match_tensor_sweep_over is True:
                logger.info(
                    "ref/base domain vs each domain match \
                            traversed one sweep, starting new epoch"
                )
                self.flag_match_tensor_sweep_over = False
                break

        if epoch < self.aconf.epochs_ctr:
            logger.info("\n\nPhase ctr-only continue\n\n")
            self.observer.reset()
            return False

        logger.info("\n\nPhase erm+ctr \n\n")
        self.flag_erm = True
        flag_stop = self.observer.update(epoch)  # notify observer
        return flag_stop

    def tr_batch(self, epoch, batch_idx, x_e, y_e, d_e, others=None):
        """
        update network for each batch
        """
        self.optimizer.zero_grad()
        x_e = x_e.to(self.device)  # 64 * 1 * 224 * 224
        # y_e_scalar = torch.argmax(y_e, dim=1).to(self.device)
        y_e = y_e.to(self.device)
        # d_e = torch.argmax(d_e, dim=1).numpy()
        d_e = d_e.to(self.device)
        # for each batch, the list loss is re-initialized

        # CTR (contrastive) loss for CTR/ERM phase are different
        list_batch_loss_ctr = []
        # for a single batch,  loss need to be
        # aggregated across different combinations of domains.
        # Defining a leaf node can cause problem by loss_ctr += xxx,
        # a list with python built-in "sum" can aggregate
        # these losses within one batch

        if self.flag_erm:
            # decoratee can be both trainer or model
            list_loss_reg_rand, list_mu_reg = self.decoratee.cal_reg_loss(
                x_e, y_e, d_e, others
            )
            loss_reg = self.model.list_inner_product(list_loss_reg_rand, list_mu_reg)
            loss_task_rand = self.model.cal_task_loss(x_e, y_e)
            # loss_erm_rnd_loader, *_ = self.model.cal_loss(x_e, y_e, d_e, others)
            loss_erm_rnd_loader = (
                loss_reg + loss_task_rand * self.model.multiplier4task_loss
            )

        num_batches_match_tensor = len(self.tuple_tensor_refdomain2each)

        if batch_idx >= num_batches_match_tensor:
            self.flag_match_tensor_sweep_over = True
            return

        curr_batch_size = self.tuple_tensor_refdomain2each[batch_idx].shape[0]

        batch_tensor_ref_domain2each = self.tuple_tensor_refdomain2each[batch_idx].to(
            self.device
        )
        # make order 5 tensor: (ref_domain, domain, channel, img_h, img_w)
        # with first dimension as batch size

        # clamp the first two dimensions so the model network could map image to feature
        batch_tensor_ref_domain2each = match_tensor_reshape(
            batch_tensor_ref_domain2each
        )
        # now batch_tensor_ref_domain2each first dim will not be batch_size!
        # batch_tensor_ref_domain2each.shape torch.Size([40, channel, 224, 224])

        batch_feat_ref_domain2each = self.model.extract_semantic_feat(
            batch_tensor_ref_domain2each
        )
        # batch_feat_ref_domain2each.shape torch.Size[40, 512]
        # torch.sum(torch.isnan(batch_tensor_ref_domain2each))
        # assert not torch.sum(torch.isnan(batch_feat_ref_domain2each))
        flag_isnan = torch.any(torch.isnan(batch_feat_ref_domain2each))
        if flag_isnan:
            logger = Logger.get_logger()
            logger.info(batch_tensor_ref_domain2each)
            raise RuntimeError(
                "batch_feat_ref_domain2each NAN! is learning rate too big or"
                "hyper-parameter tau not set appropriately?"
            )

        # for contrastive training phase,
        # the last layer of the model is replaced with identity

        batch_ref_domain2each_y = self.tuple_tensor_ref_domain2each_y[batch_idx].to(
            self.device
        )
        batch_ref_domain2each_y = batch_ref_domain2each_y.view(
            batch_ref_domain2each_y.shape[0] * batch_ref_domain2each_y.shape[1]
        )

        if self.flag_erm:
            # @FIXME: check if batch_ref_domain2each_y is
            # continuous number which means it is at its initial value,
            # not yet filled
            loss_erm_match_tensor, *_ = self.model.cal_task_loss(
                batch_tensor_ref_domain2each, batch_ref_domain2each_y.long()
            )

        # Creating tensor of shape (domain size, total domains, feat size )
        # The match tensor's first two dimension
        # [(Ref domain size) * (# train domains)]
        # has been clamped together to get features extracted
        # through self.model

        # it has to be reshaped into the match tensor shape, the same
        # for the extracted feature here, it has to reshaped into
        # the shape of the match tensor
        # to make sure that the reshape only happens at the
        # first two dimension, the feature dim has to be kept intact
        dim_feat = batch_feat_ref_domain2each.shape[1]
        num_domain_tr = len(self.task.list_domain_tr)
        batch_feat_ref_domain2each = batch_feat_ref_domain2each.view(
            curr_batch_size, num_domain_tr, dim_feat
        )

        batch_ref_domain2each_y = batch_ref_domain2each_y.view(
            curr_batch_size, num_domain_tr
        )

        # The match tensor's first two dimension
        # [(Ref domain size) * (# train domains)] has been clamped
        # together to get features extracted through self.model
        batch_tensor_ref_domain2each = batch_tensor_ref_domain2each.view(
            curr_batch_size,
            num_domain_tr,
            batch_tensor_ref_domain2each.shape[1],  # channel
            batch_tensor_ref_domain2each.shape[2],  # img_h
            batch_tensor_ref_domain2each.shape[3],
        )  # img_w

        # Contrastive Loss: class \times domain \times domain
        counter_same_cls_diff_domain = 1
        logger = Logger.get_logger()
        for y_c in range(self.task.dim_y):
            subset_same_cls = batch_ref_domain2each_y[:, 0] == y_c
            subset_diff_cls = batch_ref_domain2each_y[:, 0] != y_c
            feat_same_cls = batch_feat_ref_domain2each[subset_same_cls]
            feat_diff_cls = batch_feat_ref_domain2each[subset_diff_cls]
            logger.debug(
                f"class {y_c} with same class and different class: "
                + f"{feat_same_cls.shape[0]} {feat_diff_cls.shape[0]}"
            )

            if feat_same_cls.shape[0] == 0 or feat_diff_cls.shape[0] == 0:
                logger.debug(
                    f"no instances of label {y_c}" f"in the current batch, continue"
                )
                continue

            if torch.sum(torch.isnan(feat_diff_cls)):
                raise RuntimeError("feat_diff_cls has nan entrie(s)")

            feat_diff_cls = feat_diff_cls.view(
                feat_diff_cls.shape[0] * feat_diff_cls.shape[1], feat_diff_cls.shape[2]
            )

            for d_i in range(feat_same_cls.shape[1]):
                dist_diff_cls_same_domain = dist_pairwise_cosine(
                    feat_same_cls[:, d_i, :], feat_diff_cls[:, :]
                )

                if torch.sum(torch.isnan(dist_diff_cls_same_domain)):
                    raise RuntimeError("dist_diff_cls_same_domain NAN")

                # iterate other domains
                for d_j in range(feat_same_cls.shape[1]):
                    if d_i >= d_j:
                        continue
                    dist_same_cls_diff_domain = dist_cosine_agg(
                        feat_same_cls[:, d_i, :], feat_same_cls[:, d_j, :]
                    )

                    if torch.sum(torch.isnan(dist_same_cls_diff_domain)):
                        raise RuntimeError("dist_same_cls_diff_domain NAN")

                    # CTR (contrastive) loss is exclusive for
                    # CTR phase and ERM phase

                    if self.flag_erm:
                        list_batch_loss_ctr.append(torch.sum(dist_same_cls_diff_domain))
                    else:
                        i_dist_same_cls_diff_domain = 1.0 - dist_same_cls_diff_domain
                        i_dist_same_cls_diff_domain = (
                            i_dist_same_cls_diff_domain / self.aconf.tau
                        )
                        partition = torch.log(
                            torch.exp(i_dist_same_cls_diff_domain)
                            + dist_diff_cls_same_domain
                        )
                        list_batch_loss_ctr.append(
                            -1 * torch.sum(i_dist_same_cls_diff_domain - partition)
                        )

                    counter_same_cls_diff_domain += dist_same_cls_diff_domain.shape[0]

        loss_ctr = g_list_loss_agg(list_batch_loss_ctr) / counter_same_cls_diff_domain

        if self.flag_erm:
            epos = self.aconf.epos
        else:
            epos = self.aconf.epochs_ctr
        percentage_finished_epochs = (epoch + 1) / (epos + 1)
        # loss aggregation is over different domain
        # combinations of the same batch
        # https://discuss.pytorch.org/t/leaf-variable-was-used-in-an-inplace-operation/308
        # Loosely, tensors you create directly are leaf variables.
        # Tensors that are the result of a differentiable operation are
        # not leaf variables

        if self.flag_erm:
            # extra loss of ERM phase: the ERM loss
            # (the CTR loss for the ctr phase and erm phase are different)
            # erm loss comes from two different data loaders,
            # one is rnd (random) data loader
            # the other one is the data loader from the match tensor
            loss_e = (
                torch.tensor(0.0, requires_grad=True)
                + g_inst_component_loss_agg(loss_erm_rnd_loader)
                + g_inst_component_loss_agg(loss_erm_match_tensor)
                * self.model.multiplier4task_loss
                + self.lambda_ctr * percentage_finished_epochs * loss_ctr
            )
        else:
            loss_e = (
                torch.tensor(0.0, requires_grad=True)
                + self.lambda_ctr * percentage_finished_epochs * loss_ctr
            )
        # @FIXME: without torch.tensor(0.0), after a few epochs,
        # error "'float' object has no attribute 'backward'"

        loss_e.backward(retain_graph=False)
        self.optimizer.step()
        self.epo_loss_tr += loss_e.detach().item()

        torch.cuda.empty_cache()

    def mk_match_tensor(self, epoch):
        """
        initialize or update match tensor
        """
        obj_match = MatchPair(
            self.task.dim_y,
            self.task.isize.i_c,
            self.task.isize.i_h,
            self.task.isize.i_w,
            self.aconf.bs,
            virtual_ref_dset_size=self.base_domain_size,
            num_domains_tr=len(self.task.list_domain_tr),
            list_tr_domain_size=self.list_tr_domain_size,
        )

        # @FIXME: what is the usefulness of (epoch > 0) as argument
        (
            self.tensor_ref_domain2each_domain_x,
            self.tensor_ref_domain2each_domain_y,
        ) = obj_match(
            self.device,
            self.task.loader_tr,
            self.model.extract_semantic_feat,
            (epoch > 0),
        )

    def before_tr(self):
        """
        override abstract method
        """
        logger = Logger.get_logger()
        logger.info("\n\nPhase 1 start: contractive alignment without task loss: \n\n")
        # phase 1: contrastive learning
        # different than phase 2, ctr_model has no classification loss


def match_tensor_reshape(batch_tensor_ref_domain2each):
    """
    # original dimension is (ref_domain, domain, (channel, img_h, img_w))
    # use a function so it is easier to accomodate other data mode (not image)
    """
    batch_tensor_refdomain_other_domain_chw = batch_tensor_ref_domain2each.view(
        batch_tensor_ref_domain2each.shape[0] * batch_tensor_ref_domain2each.shape[1],
        batch_tensor_ref_domain2each.shape[2],  # channel
        batch_tensor_ref_domain2each.shape[3],  # img_h
        batch_tensor_ref_domain2each.shape[4],
    )  # img_w
    return batch_tensor_refdomain_other_domain_chw



"""
use random start to generate adversarial images
"""
import torch
from torch.autograd import Variable

from domainlab.algos.trainers.train_basic import TrainerBasic


class TrainerDIAL(TrainerBasic):
    """
    Trainer Domain Invariant Adversarial Learning
    """

    def gen_adversarial(self, device, img_natural, vec_y):
        """
        use naive trimming to find optimize img in the direction of adversarial gradient,
        this is not necessarily constraint optimal due to nonlinearity,
        as the constraint epsilon is only considered ad-hoc
        """
        # ensure adversarial image not in computational graph
        steps_perturb = self.aconf.dial_steps_perturb
        scale = self.aconf.dial_noise_scale
        step_size = self.aconf.dial_lr
        epsilon = self.aconf.dial_epsilon
        img_adv_ini = img_natural.detach()
        img_adv_ini = (
            img_adv_ini + scale * torch.randn(img_natural.shape).to(device).detach()
        )
        img_adv = img_adv_ini
        for _ in range(steps_perturb):
            img_adv.requires_grad_()
            loss_gen_adv = self.model.cal_loss_gen_adv(img_natural, img_adv, vec_y)
            grad = torch.autograd.grad(loss_gen_adv, [img_adv])[0]
            # instead of gradient descent, we gradient ascent here
            img_adv = img_adv_ini.detach() + step_size * torch.sign(grad.detach())
            img_adv = torch.min(
                torch.max(img_adv, img_natural - epsilon), img_natural + epsilon
            )
            img_adv = torch.clamp(img_adv, 0.0, 1.0)
        return img_adv

    def _cal_reg_loss(self, tensor_x, tensor_y, tensor_d, others=None):
        """
        Let trainer behave like a model, so that other trainer could use it
        """
        _ = tensor_d
        _ = others
        tensor_x_adv = self.gen_adversarial(self.device, tensor_x, tensor_y)
        tensor_x_batch_adv_no_grad = Variable(tensor_x_adv, requires_grad=False)
        loss_dial = self.model.cal_task_loss(tensor_x_batch_adv_no_grad, tensor_y)
        return [loss_dial], [self.aconf.gamma_reg]



"""
observer and visitor pattern, responsible train, validation, test
dispatch performance evaluation to model,
dispatch model selection to model selection object
"""
import os

from domainlab.algos.observers.a_observer import AObVisitor
from domainlab.tasks.task_folder_mk import NodeTaskFolderClassNaMismatch
from domainlab.tasks.task_pathlist import NodeTaskPathListDummy
from domainlab.utils.logger import Logger


class ObVisitor(AObVisitor):
    """
    Observer + Visitor pattern for model selection
    """

    def __init__(self, model_sel):
        """
        observer trainer
        """
        super().__init__()
        self.host_trainer = None
        self.model_sel = model_sel
        self.epo = None
        self.metric_te = None
        self.metric_val = None
        self.perf_metric = None

    @property
    def str_metric4msel(self):
        """
        string representing the metric used for persisting models on the disk
        """
        return self.host_trainer.str_metric4msel

    def update(self, epoch):
        logger = Logger.get_logger()
        logger.info(f"epoch: {epoch}")
        self.epo = epoch
        if epoch % self.epo_te == 0:
            logger.info("---- Training Domain: ")
            self.host_trainer.model.cal_perf_metric(self.loader_tr, self.device)
            if self.loader_val is not None:
                logger.info("---- Validation: ")
                self.metric_val = self.host_trainer.model.cal_perf_metric(
                    self.loader_val, self.device
                )
            if self.loader_te is not None:
                logger.info("---- Test Domain (oracle): ")
                metric_te = self.host_trainer.model.cal_perf_metric(
                    self.loader_te, self.device
                )
                self.metric_te = metric_te
        if self.model_sel.update():
            logger.info("better model found")
            self.host_trainer.model.save()
            logger.info("persisted")
        flag_stop = self.model_sel.if_stop()
        flag_enough = epoch >= self.host_trainer.aconf.epos_min
        return flag_stop & flag_enough

    def accept(self, trainer):
        """
        accept invitation as a visitor
        """
        self.host_trainer = trainer
        self.model_sel.accept(trainer, self)
        self.set_task(trainer.task, args=trainer.aconf, device=trainer.device)
        self.perf_metric = self.host_trainer.model.create_perf_obj(self.task)

    def after_all(self):
        """
        After training is done
        """
        model_ld = None
        try:
            model_ld = self.host_trainer.model.load()
        except FileNotFoundError as err:
            # if other errors/exceptions occur, we do not catch them
            # other exceptions will terminate the python script
            # this can happen if loss is increasing, model never get selected
            logger = Logger.get_logger()
            logger.warning(err)
            logger.warning(
                "this error can occur if model selection criteria \
                           is worsening, "
                "model never get persisted, \
                           no performance metric is reported"
            )
            return
        model_ld = model_ld.to(self.device)
        model_ld.eval()
        logger = Logger.get_logger()
        logger.info("persisted model performance metric: \n")
        metric_te = model_ld.cal_perf_metric(self.loader_te, self.device)
        dict_2add = self.cal_oracle_perf()
        if dict_2add is not None:
            metric_te.update(dict_2add)
        else:
            metric_te.update({"acc_oracle": -1})
        if hasattr(self, "model_sel"):
            metric_te.update({"acc_val": self.model_sel.best_val_acc})
        else:
            metric_te.update({"acc_val": -1})
        self.dump_prediction(model_ld, metric_te)
        # save metric to one line in csv result file
        self.host_trainer.model.visitor(metric_te)
        # prediction dump of test domain is essential
        # to verify the prediction results

    def cal_oracle_perf(self):
        """
        calculate oracle performance
        """
        try:
            model_or = self.host_trainer.model.load("oracle")
            # @FIXME: name "oracle is a strong dependency
            model_or = model_or.to(self.device)
            model_or.eval()
        except FileNotFoundError:
            return {"acc_oracle": -1}
        logger = Logger.get_logger()
        logger.info("oracle model performance metric: \n")
        metric_te = model_or.cal_perf_metric(self.loader_te, self.device)
        return {"acc_oracle": metric_te["acc"]}

    def dump_prediction(self, model_ld, metric_te):
        """
        given the test domain loader, use the loaded model \
            model_ld to predict each instance
        """
        flag_task_folder = isinstance(
            self.host_trainer.task, NodeTaskFolderClassNaMismatch
        )
        flag_task_path_list = isinstance(self.host_trainer.task, NodeTaskPathListDummy)
        if flag_task_folder or flag_task_path_list:
            fname4model = (
                self.host_trainer.model.visitor.model_path
            )  # pylint: disable=E1101
            file_prefix = os.path.splitext(fname4model)[0]  # remove ".model"
            dir4preds = os.path.join(self.host_trainer.aconf.out, "saved_predicts")
            if not os.path.exists(dir4preds):
                os.mkdir(dir4preds)
            file_prefix = os.path.join(dir4preds, os.path.basename(file_prefix))
            file_name = file_prefix + "_instance_wise_predictions.txt"
            model_ld.pred2file(
                self.loader_te, self.device, filename=file_name, metric_te=metric_te
            )

    def clean_up(self):
        """
        to be called by a decorator
        """
        if not self.keep_model:
            self.host_trainer.model.visitor.clean_up()



"""
parent class for combing model, trainer, task, observer
"""
import abc

from domainlab.compos.pcr.p_chain_handler import AbstractChainNodeHandler
from domainlab.utils.logger import Logger


class NodeAlgoBuilder(AbstractChainNodeHandler):
    """
    Base class for Algorithm Builder
    """

    na_prefix = "NodeAlgoBuilder"

    @property
    def name(self):
        """
        get the name of the algorithm
        """
        na_prefix = NodeAlgoBuilder.na_prefix
        len_prefix = len(na_prefix)
        na_class = type(self).__name__
        if na_class[:len_prefix] != na_prefix:
            raise RuntimeError(
                "algorithm builder node class must start with ",
                na_prefix,
                "the current class is named: ",
                na_class,
            )
        return type(self).__name__[len_prefix:].lower()

    def is_myjob(self, request):
        """
        :param request: string
        """
        return request == self.name

    @abc.abstractmethod
    def init_business(self, exp):
        """
        combine model, trainer, observer, task
        """

    def extend(self, node):
        """
        Extends the current algorithm builder with a new node.

        This method updates the builder by setting the `next_model` attribute to the specified node.

        Args:
            node: The node to be added to the algorithm builder.
        """
        self.next_model = node

    def init_next_model(self, model, exp):
        """
        initialize the next model and decorate with current model
        """
        if self.next_model is not None:
            _, next_model, *_ = self.next_model.init_business(exp)
            model.extend(next_model)
        return model



"""
Integrate Chain-of-Responsibility and Builder Patter to construct VAE encoder and decoder
"""

import abc

from domainlab.compos.pcr.p_chain_handler import AbstractChainNodeHandler


class AbstractModelBuilderChainNode(AbstractChainNodeHandler):
    """
    to ensure chain of responsibility node AbstractChainNodeHandler always
    work even some node can not start their heavy weight business object, avoid override the
    initializer so that node construction is always light weight.

    The config() method here is abstract, while child class has a concrete config method
    """

    @abc.abstractmethod
    def config(self, *kargs, **kwargs):
        """
        use either list or dictionary input arguments to configure the model builder.
        """
        raise NotImplementedError



"""
This class is used to perform the sanity check on a task description
"""

import datetime
import os
import shutil

import numpy as np
import torch.utils.data as data_utils
from torch.utils.data import Subset

from domainlab.dsets.utils_data import plot_ds


class SanityCheck:
    """
    Performs a sanity check on the given args and the task
    when running dataset_sanity_check(self)
    """

    def __init__(self, args, task):
        self.args = args
        self.task = task

    def dataset_sanity_check(self):
        """
        when we load data from folder or a file listing the path of observations,
        we want to check if the file we loaded are in accordance with our expectations
        This function dump a subsample of the dataset into hierarchical folder structure.
        """
        # self.task.init_business(self.args)

        list_domain_tr, list_domain_te = self.task.get_list_domains_tr_te(
            self.args.tr_d, self.args.te_d
        )

        time_stamp = datetime.datetime.now()
        f_name = os.path.join(
            self.args.out,
            "Dset_extraction",
            self.task.task_name + " " + str(time_stamp),
        )
        # remove previous sanity checks with the same name
        shutil.rmtree(f_name, ignore_errors=True)

        # for each training domain do...
        for domain in list_domain_tr:
            if domain in self.task.dict_dset_tr:
                d_dataset = self.task.dict_dset_tr[domain]
            else:
                d_dataset = self.task.get_dset_by_domain(self.args, domain)[0]
            folder_name = f_name + "/train_domain/" + str(domain)
            self.save_san_check_for_domain(self.args.san_num, folder_name, d_dataset)

        # for each testing domain do...
        for domain in list_domain_te:
            if domain in self.task.dict_dset_te:
                d_dataset = self.task.dict_dset_te[domain]
            else:
                d_dataset = self.task.get_dset_by_domain(self.args, domain)[0]
            folder_name = f_name + "/test_domain/" + str(domain)
            self.save_san_check_for_domain(self.args.san_num, folder_name, d_dataset)

    def save_san_check_for_domain(self, sample_num, folder_name, d_dataset):
        """
        saves a extraction of the dataset (d_dataset) into folder (folder_name)
        sample_num: int, number of images which are extracted from the dataset
        folder_name: string, destination for the saved images
        d_dataset: dataset
        """

        # for each class do...
        for class_num in range(len(self.task.list_str_y)):
            num_of_samples = 0
            loader_domain = data_utils.DataLoader(
                d_dataset, batch_size=1, shuffle=False
            )
            domain_targets = []
            for num, (_, lab, *_) in enumerate(loader_domain):
                if int(np.argmax(lab[0])) == class_num:
                    domain_targets.append(num)
                    num_of_samples += 1
                if sample_num == num_of_samples:
                    break

            class_dataset = Subset(d_dataset, domain_targets)
            os.makedirs(folder_name, exist_ok=True)
            plot_ds(
                class_dataset,
                folder_name + "/" + str(self.task.list_str_y[class_num]) + ".jpg",
                batchsize=sample_num,
            )



import torch
import torch.nn as nn
from torch.distributions import Dirichlet


class EncoderH2Dirichlet(nn.Module):
    """
    hidden representation to Dirichlet Distribution
    """

    def __init__(self, dim_topic, device):
        """ """
        super().__init__()
        self.layer_bn = nn.BatchNorm1d(dim_topic)
        self.layer_concentration = nn.Softplus()
        self.device = device

    def forward(self, hidden):
        """
        :param hidden:
        """
        feat_bnorm = self.layer_bn(hidden)
        # alphas_batch = torch.log(1 + feat_bnorm.exp())
        alphas_batch = self.layer_concentration(feat_bnorm)
        q_topic = Dirichlet(alphas_batch + 1e-6)
        topic_q = q_topic.rsample().to(self.device)
        return q_topic, topic_q



from torch import nn
from torch.autograd import Function


class Flatten(nn.Module):
    def forward(self, x):
        x = x.view(x.size()[0], -1)
        return x


class AutoGradFunReverseMultiply(Function):
    """
    https://pytorch.org/docs/stable/autograd.html
    https://pytorch.org/docs/stable/notes/extending.html#extending-autograd
    """

    @staticmethod
    def forward(ctx, x, alpha):
        ctx.alpha = alpha
        return x.view_as(x)

    @staticmethod
    def backward(ctx, grad_output):
        output = grad_output.neg() * ctx.alpha
        return output, None


class AutoGradFunMultiply(Function):
    @staticmethod
    def forward(ctx, x, alpha):
        ctx.alpha = alpha
        return x.view_as(x)

    @staticmethod
    def backward(ctx, grad_output):
        output = grad_output * ctx.alpha
        return output, None



"""
Chain node VAE builders
"""
from domainlab.compos.vae.c_vae_builder_classif import (
    ChainNodeVAEBuilderClassifCondPrior,
)
from domainlab.compos.vae.compos.decoder_concat_vec_reshape_conv_gated_conv import (
    DecoderConcatLatentFCReshapeConvGatedConv,
)
from domainlab.compos.vae.compos.encoder_xyd_parallel import (
    XYDEncoderParallelAlex,
    XYDEncoderParallelConvBnReluPool,
    XYDEncoderParallelExtern,
    XYDEncoderParallelUser,
)


class ChainNodeVAEBuilderClassifCondPriorBase(ChainNodeVAEBuilderClassifCondPrior):
    """
    base class of AE builder
    """

    def config_img(self, flag, request):
        """config_img.

        :param flag:
        :param request:
        """
        if flag:
            self.i_c = request.i_c
            self.i_h = request.i_h
            self.i_w = request.i_w

    def is_myjob(self, request):
        """is_myjob.

        :param request:
        """
        raise NotImplementedError

    def build_encoder(self):
        """build_encoder."""
        raise NotImplementedError

    def build_decoder(self):
        """build_decoder."""
        decoder = DecoderConcatLatentFCReshapeConvGatedConv(
            z_dim=self.zd_dim + self.zx_dim + self.zy_dim,
            i_c=self.i_c,
            i_w=self.i_w,
            i_h=self.i_h,
        )
        return decoder


class NodeVAEBuilderArg(ChainNodeVAEBuilderClassifCondPriorBase):
    """Build encoder decoder according to commandline arguments"""

    def is_myjob(self, request):
        """is_myjob.
        :param request:
        """
        self.request = request
        self.args = request.args
        self.config_img(True, request)
        if self.args.npath is not None or self.args.npath_dom is not None:
            return True
        return False

    def build_encoder(self):
        """build_encoder."""
        encoder = XYDEncoderParallelExtern(
            self.zd_dim,
            self.zx_dim,
            self.zy_dim,
            args=self.args,
            i_c=self.i_c,
            i_h=self.i_h,
            i_w=self.i_w,
        )
        return encoder


class NodeVAEBuilderUser(ChainNodeVAEBuilderClassifCondPriorBase):
    """Build encoders according to test_mk_exp file"""

    def is_myjob(self, request):
        flag = not hasattr(request, "args")
        self.request = request
        self.config_img(flag, request)
        return flag

    def build_encoder(self):
        encoder = XYDEncoderParallelUser(
            self.request.net_class_d, self.request.net_x, self.request.net_class_y
        )
        return encoder


class NodeVAEBuilderImgConvBnPool(ChainNodeVAEBuilderClassifCondPriorBase):
    def is_myjob(self, request):
        """is_myjob.

        :param request:
        """
        flag = (
            request.args.nname == "conv_bn_pool_2"
            or request.args.nname_dom == "conv_bn_pool_2"
        )  # @FIXME
        self.config_img(flag, request)
        return flag

    def build_encoder(self):
        """build_encoder."""
        encoder = XYDEncoderParallelConvBnReluPool(
            self.zd_dim, self.zx_dim, self.zy_dim, self.i_c, self.i_h, self.i_w
        )
        return encoder


class NodeVAEBuilderImgAlex(NodeVAEBuilderImgConvBnPool):
    """NodeVAEBuilderImgAlex"""

    def is_myjob(self, request):
        """is_myjob.

        :param request:
        """
        self.args = request.args
        flag = self.args.nname == "alexnet"  # @FIXME
        self.config_img(flag, request)
        return flag

    def build_encoder(self):
        """build_encoder."""
        encoder = XYDEncoderParallelAlex(
            self.zd_dim,
            self.zx_dim,
            self.zy_dim,
            self.i_c,
            self.i_h,
            self.i_w,
            args=self.args,
        )
        return encoder



"""
Use dictionaries to create train and test domain split
"""
import torch

from domainlab.tasks.a_task_classif import NodeTaskDGClassif
from domainlab.tasks.b_task import NodeTaskDict


class NodeTaskDictClassif(NodeTaskDict, NodeTaskDGClassif):
    """
    Use dictionaries to create train and test domain split
    """

    def init_business(self, args, trainer=None):
        """
        create a dictionary of datasets
        """
        super().init_business(args=args, trainer=trainer)
        self.count_domain_class()

    def count_domain_class(self):
        """
        iterate all domains and count the class label distribution for each
        return a double dictionary {"domain1": {"class1":3, "class2": 4,...}, ....}
        """
        for key, dset in self.dict_dset_tr.items():
            dict_class_count = self._count_class_one_hot(dset)
            self.dict_domain_class_count[key] = dict_class_count
        for key, dset in self.dict_dset_te.items():
            dict_class_count = self._count_class_one_hot(dset)
            self.dict_domain_class_count[key] = dict_class_count

    def _count_class_one_hot(self, dset):
        labels_count = torch.zeros(self.dim_y, dtype=torch.long)
        for _, target, *_ in dset:
            labels_count += target.long()

        list_count = list(labels_count.cpu().numpy())
        dict_class_count = {}
        for name, count in zip(self.list_str_y, list_count):
            dict_class_count[name] = count
        return dict_class_count



"""
Color MNIST with palette
"""
from torch.utils.data import random_split

from domainlab.dsets.dset_mnist_color_solo_default import DsetMNISTColorSoloDefault
from domainlab.dsets.utils_color_palette import default_rgb_palette  # @FIXME
from domainlab.tasks.b_task_classif import NodeTaskDictClassif
from domainlab.tasks.utils_task import ImSize
from domainlab.utils.utils_classif import mk_dummy_label_list_str


class NodeTaskMNISTColor10(NodeTaskDictClassif):
    """
    Use the deafult palette with 10 colors
    """

    @property
    def list_str_y(self):
        return mk_dummy_label_list_str("digit", 10)

    @property
    def isize(self):
        """image channel, height, width"""
        return ImSize(3, 28, 28)

    def get_list_domains(self):
        """
        1. get list of domain names
        2. better use method than property so new domains can be added
        """
        list_domains = []
        for rgb_list in default_rgb_palette:  # 10 colors
            domain = "_".join([str(c) for c in rgb_list])
            domain = "rgb_" + domain
            list_domains.append(domain)
        return list_domains

    def get_dset_by_domain(
        self, args, na_domain, split=True
    ):  # @FIXME: different number of arguments than parent
        """get_dset_by_domain.
        :param args:
        :param na_domain:
        :param split: for test set, no need to split
        args.split: by default, split is set to be zero which in python can
        be evaluated in if statement, in which case, no validation set will be
        created. Otherwise, this argument is the split ratio
        """
        ratio_split = float(args.split) if split else False
        # by default, split is set to be zero which in python can
        # be evaluated in if statement, in which case, no validation
        # set will be created. Otherwise, this argument is
        # the split ratio
        ind_global = self.get_list_domains().index(na_domain)
        dset = DsetMNISTColorSoloDefault(ind_global, args.dpath)
        train_set = dset
        val_set = dset
        # split dset into training and test
        if ratio_split:
            train_len = int(len(dset) * ratio_split)
            val_len = len(dset) - train_len
            train_set, val_set = random_split(dset, [train_len, val_len])
        return train_set, val_set



import torch


def get_flat_dim(module, i_channel, i_h, i_w, batchsize=5):
    """flat the convolution layer output and get the flat dimension for fully
    connected network
    :param module:
    :param i_channel:
    :param i_h:
    :param i_w:
    :param batchsize:
    """
    img = torch.randn(i_channel, i_h, i_w)
    img3 = img.repeat(batchsize, 1, 1, 1)  # create batchsize repitition
    conv_output = module(img3)
    if len(conv_output.shape) == 2:
        flat_dim = conv_output.shape[1]
    else:
        flat_dim = conv_output.shape[1] * conv_output.shape[2] * conv_output.shape[3]
    return flat_dim



from domainlab.dsets.a_dset_mnist_color_rgb_solo import ADsetMNISTColorRGBSolo
from domainlab.dsets.utils_color_palette import default_rgb_palette


class DsetMNISTColorSoloDefault(ADsetMNISTColorRGBSolo):
    @property
    def palette(self):
        return default_rgb_palette

    def get_num_colors(self):
        return len(self.palette)

    def get_background_color(self, ind):
        if self.color_scheme == "back":
            return self.palette[ind]
        if self.color_scheme == "both":
            return self.palette[-(ind - 3)]
        # only array can be multiplied with number 255 directly
        return self.palette[ind]
        # "num" do not use background at all

    def get_foreground_color(self, ind):
        if self.color_scheme == "num":
            return self.palette[-(ind + 1)]
        return self.palette[ind]



import torch.nn as nn


class GatedDense(nn.Module):
    def __init__(self, input_size, output_size, activation=None):
        super(GatedDense, self).__init__()

        self.activation = activation
        self.sigmoid = nn.Sigmoid()
        self.h = nn.Linear(input_size, output_size)
        self.g = nn.Linear(input_size, output_size)

    def forward(self, x):
        h = self.h(x)
        if self.activation is not None:
            h = self.activation(self.h(x))

        g = self.sigmoid(self.g(x))

        return h * g


# ==========================================================================
class GatedConv2d(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        kernel_size,
        stride,
        padding,
        dilation=1,
        activation=None,
    ):
        super(GatedConv2d, self).__init__()

        self.activation = activation
        self.sigmoid = nn.Sigmoid()

        self.h = nn.Conv2d(
            input_channels, output_channels, kernel_size, stride, padding, dilation
        )
        self.g = nn.Conv2d(
            input_channels, output_channels, kernel_size, stride, padding, dilation
        )

    def forward(self, x):
        if self.activation is None:
            h = self.h(x)
        else:
            h = self.activation(self.h(x))

        g = self.sigmoid(self.g(x))

        return h * g


# ==============================================================================
class Conv2d(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        kernel_size,
        stride,
        padding,
        dilation=1,
        activation=None,
        bias=True,
    ):
        super(Conv2d, self).__init__()

        self.activation = activation
        self.conv = nn.Conv2d(
            input_channels,
            output_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            bias=bias,
        )

    def forward(self, x):
        h = self.conv(x)
        if self.activation is None:
            out = h
        else:
            out = self.activation(h)
        return out



"""
Integrate Chain-of-Responsibility and Builder Patter to construct VAE
encoder and decoder
"""

import abc

from domainlab.compos.pcr.p_chain_handler import AbstractChainNodeHandler
from domainlab.utils.utils_class import store_args


class AbstractVAEBuilderChainNode(AbstractChainNodeHandler):
    """
    to ensure chain of responsibility node AbstractChainNodeHandler always
    work even some node can not start their heavy weight business object,
    avoid override the
    initializer so that node construction is always light weight.
    """

    def __init__(self, successor_node):
        self.args = None
        self.zd_dim = None
        self.zx_dim = None
        self.zy_dim = None
        self.i_c = None
        self.i_h = None
        self.i_w = None
        super().__init__(successor_node)

    @store_args
    def init_business(self, zd_dim, zx_dim, zy_dim):
        """
        initialize **and** return the heavy weight business object for doing the real job
        :param request: subclass can override request object to be string or function
        :return: the constructed service object
        """
        return self

    @abc.abstractmethod
    def build_encoder(self):
        raise NotImplementedError

    @abc.abstractmethod
    def build_decoder(self):
        raise NotImplementedError



"""
merge several solo-color mnist to form a mixed dataset
"""
import numpy as np
from torch.utils.data import Dataset

from domainlab.dsets.dset_mnist_color_solo_default import DsetMNISTColorSoloDefault
from domainlab.dsets.utils_data import mk_fun_label2onehot


class DsetMNISTColorMix(Dataset):
    """
    merge several solo-color mnist to form a mixed dataset
    """

    def __init__(self, n_domains, path, color_scheme="both"):
        self.n_domains = n_domains
        self.list_dset = [None] * n_domains
        self.fun_dlabel2onehot = mk_fun_label2onehot(n_domains)
        for domain_ind in range(n_domains):
            self.list_dset[domain_ind] = DsetMNISTColorSoloDefault(
                domain_ind, path, color_scheme=color_scheme
            )
        self.list_len = [len(ds) for ds in self.list_dset]
        self.size_single = min(self.list_len)

    def __len__(self):
        """__len__."""
        return sum(self.list_len)

    def __getitem__(self, idx):
        rand_domain = np.random.random_integers(self.n_domains - 1)  # @FIXME
        idx_local = idx % self.size_single
        img, c_label = self.list_dset[rand_domain][idx_local]
        return img, c_label, self.fun_dlabel2onehot(rand_domain)


class DsetMNISTColorMixNoDomainLabel(DsetMNISTColorMix):
    """
    DsetMNISTColorMixNoDomainLabel
    """

    def __getitem__(self, idx):
        img, c_label, _ = super().__getitem__(idx)
        return img, c_label



import torch.nn as nn

from domainlab.utils.logger import Logger


class NetTorchVisionBase(nn.Module):
    """
    fetch model from torchvision
    """

    def __init__(self, flag_pretrain):
        super().__init__()
        self.net_torchvision = None
        self.fetch_net(flag_pretrain)

    def fetch_net(self, flag_pretrain):
        raise NotImplementedError

    def forward(self, tensor):
        """
        delegate forward operation to self.net_torchvision
        """
        out = self.net_torchvision(tensor)
        return out

    def show(self):
        """
        print out which layer will be optimized
        """
        for name, param in self.net_torchvision.named_parameters():
            if param.requires_grad:
                logger = Logger.get_logger()
                logger.info(f"layers that will be optimized: \t{name}")



import warnings

import numpy as np
import torch

from domainlab.algos.trainers.compos.matchdg_utils import (
    MatchDictNumDomain2SizeDomain,
    MatchDictVirtualRefDset2EachDomain,
)
from domainlab.tasks.utils_task import mk_loader
from domainlab.utils.logger import Logger
from domainlab.utils.utils_class import store_args


class MatchPair:
    """
    match different input
    """
    @store_args
    def __init__(
        self,
        dim_y,
        i_c,
        i_h,
        i_w,
        bs_match,
        virtual_ref_dset_size,
        num_domains_tr,
        list_tr_domain_size,
    ):
        """
        :param virtual_ref_dset_size:  sum of biggest class sizes
        :param num_domains_tr:
        :param list_tr_domain_size:
        """
        self.dim_y = dim_y
        self.bs_match = bs_match
        self.virtual_ref_dset_size = virtual_ref_dset_size
        self.num_domains_tr = num_domains_tr
        self.list_tr_domain_size = list_tr_domain_size
        self.dict_cls_ind_base_domain_ind = {}
        self.dict_virtual_dset2each_domain = MatchDictVirtualRefDset2EachDomain(
            virtual_ref_dset_size=virtual_ref_dset_size,
            num_domains_tr=num_domains_tr,
            i_c=i_c,
            i_h=i_h,
            i_w=i_w,
        )()

        self.dict_domain_data = MatchDictNumDomain2SizeDomain(
            num_domains_tr=num_domains_tr,
            list_tr_domain_size=list_tr_domain_size,
            i_c=i_c,
            i_h=i_h,
            i_w=i_w,
        )()

        self.indices_matched = {}
        for key in range(virtual_ref_dset_size):
            self.indices_matched[key] = []
        self.perfect_match_rank = []

        self.domain_count = {}
        for domain in range(num_domains_tr):
            self.domain_count[domain] = 0

    def _fill_data(self, loader):
        """
        copy all data from loader, then store them in memory
        variable self.dict_domain_data
        """
        # NOTE: loader contains data from several dataset
        list_idx_several_ds = []
        loader_full_data = mk_loader(
            loader.dataset, bsize=loader.batch_size, drop_last=False
        )
        # @FIXME: training loader will always drop the last incomplete batch
        for _, (x_e, y_e, d_e, idx_e) in enumerate(loader_full_data):
            # traverse mixed domain data from loader
            list_idx_several_ds.extend(list(idx_e.cpu().numpy()))
            y_e = torch.argmax(y_e, dim=1)
            d_e = torch.argmax(d_e, dim=1).numpy()

            # get all domains in current batch
            unique_domains = np.unique(d_e)
            for domain_idx in unique_domains:
                # select all instances belong to one domain
                flag_curr_domain = d_e == domain_idx

                # flag_curr_domain is subset indicator of
                # True of False for selection of data from the mini-batch

                # get global index of all instances of the current domain
                global_indices = idx_e[flag_curr_domain]

                # global_indices are subset of idx_e, which contains
                # global index of data from the loader
                for local_ind in range(global_indices.shape[0]):
                    # @FIXME: the following is just coping all data to
                    # self.dict_domain_data (in memory with ordering),
                    # which seems redundant

                    # tensor.item get the scalar
                    global_ind = global_indices[local_ind].item()
                    self.dict_domain_data[domain_idx]["data"][global_ind] = x_e[
                        flag_curr_domain
                    ][local_ind]
                    # flag_curr_domain are subset indicator
                    # for selection of domain
                    self.dict_domain_data[domain_idx]["label"][global_ind] = y_e[
                        flag_curr_domain
                    ][local_ind]
                    # copy trainining batch to dict_domain_data
                    self.dict_domain_data[domain_idx]["idx"][global_ind] = idx_e[
                        flag_curr_domain
                    ][local_ind]
                    self.domain_count[domain_idx] += 1

        # if all data has been re-organized(filled) into the current tensor
        assert len(list_idx_several_ds) == len(loader.dataset)
        # NOTE: check if self.dict_domain_data[domain_idx]['label'] has
        # some instances that are initial continuous
        # value instead of class label
        for domain in range(self.num_domains_tr):
            if self.domain_count[domain] != self.list_tr_domain_size[domain]:
                logger = Logger.get_logger()
                logger.warning("domain_count show matching: dictionary missing data!")
                warnings.warn("domain_count show matching: dictionary missing data!")

    def _cal_base_domain(self):
        """
        # Determine the base_domain_idx as the domain
        # with the max samples of the current class
        # Create dictionary: class label -> list of ordered flag_curr_domain
        """
        for y_c in range(self.dim_y):
            base_domain_size = 0
            base_domain_idx = -1
            for domain_idx in range(self.num_domains_tr):
                flag_curr_class = (
                    self.dict_domain_data[domain_idx]["label"] == y_c
                )  # tensor of True/False
                curr_size = self.dict_domain_data[domain_idx]["label"][
                    flag_curr_class
                ].shape[0]
                # flag_curr_class are subset indicator
                if base_domain_size < curr_size:
                    base_domain_size = curr_size
                    base_domain_idx = domain_idx
            self.dict_cls_ind_base_domain_ind[y_c] = base_domain_idx
            # for each class label, there is a base domain
            logger = Logger.get_logger()
            logger.info(f"for class {y_c}")
            logger.info(f"domain index as base domain: {base_domain_idx}")
            logger.info(f"Base Domain size {base_domain_size}")

    def __call__(self, device, loader, fun_extract_semantic_feat, flag_match_min_dist):
        """
        :param fun_extract_semantic_feat: function to generate causal features from input
        """
        self._fill_data(loader)
        self._cal_base_domain()
        for curr_domain_ind in range(self.num_domains_tr):
            counter_ref_dset_size = 0
            for y_c in range(self.dim_y):
                # which domain to use as the base domain for the current class
                base_domain_idx = self.dict_cls_ind_base_domain_ind[y_c]

                # subset indicator
                flags_base_domain_curr_cls = (
                    self.dict_domain_data[base_domain_idx]["label"] == y_c
                )
                flags_base_domain_curr_cls = flags_base_domain_curr_cls[:, 0]
                global_inds_base_domain_curr_cls = self.dict_domain_data[
                    base_domain_idx
                ]["idx"][flags_base_domain_curr_cls]
                # pick out base domain class label y_c images
                # the difference of this block is "curr_domain_ind"
                # in iteration is
                # used instead of base_domain_idx for current class

                # pick out current domain y_c class images
                flag_curr_domain_curr_cls = (
                    self.dict_domain_data[curr_domain_ind]["label"] == y_c
                )
                # NO label matches y_c
                flag_curr_domain_curr_cls = flag_curr_domain_curr_cls[:, 0]
                global_inds_curr_domain_curr_cls = self.dict_domain_data[
                    curr_domain_ind
                ]["idx"][flag_curr_domain_curr_cls]
                size_curr_domain_curr_cls = global_inds_curr_domain_curr_cls.shape[0]
                if (
                    size_curr_domain_curr_cls == 0
                ):  # there is no class y_c in current domain
                    raise RuntimeError(
                        f"current domain {curr_domain_ind} does not contain class {y_c}"
                    )

                # compute base domain features for class label y_c
                x_base_domain_curr_cls = self.dict_domain_data[base_domain_idx]["data"][
                    flags_base_domain_curr_cls
                ]
                # pick out base domain class label y_c images
                # split data into chunks
                tuple_batch_x_base_domain_curr_cls = torch.split(
                    x_base_domain_curr_cls, self.bs_match, dim=0
                )
                # @FIXME. when x_base_domain_curr_cls is smaller
                # than the self.bs_match, then there is only one batch
                list_base_feat = []
                for batch_x_base_domain_curr_cls in tuple_batch_x_base_domain_curr_cls:
                    with torch.no_grad():
                        batch_x_base_domain_curr_cls = batch_x_base_domain_curr_cls.to(
                            device
                        )
                        feat = fun_extract_semantic_feat(batch_x_base_domain_curr_cls)
                        list_base_feat.append(feat.cpu())
                tensor_feat_base_domain_curr_cls = torch.cat(list_base_feat)
                # base domain features

                if flag_match_min_dist:  # if epoch > 0:flag_match_min_dist=True
                    x_curr_domain_curr_cls = self.dict_domain_data[curr_domain_ind][
                        "data"
                    ][flag_curr_domain_curr_cls]
                    # indices_curr pick out current domain y_c class images
                    tuple_x_batch_curr_domain_curr_cls = torch.split(
                        x_curr_domain_curr_cls, self.bs_match, dim=0
                    )
                    list_feat_x_curr_domain_curr_cls = []
                    for batch_feat in tuple_x_batch_curr_domain_curr_cls:
                        with torch.no_grad():
                            batch_feat = batch_feat.to(device)
                            out = fun_extract_semantic_feat(batch_feat)
                            list_feat_x_curr_domain_curr_cls.append(out.cpu())
                    tensor_feat_curr_domain_curr_cls = torch.cat(
                        list_feat_x_curr_domain_curr_cls
                    )
                    # feature through inference network for the current domain of class y_c

                tensor_feat_base_domain_curr_cls = (
                    tensor_feat_base_domain_curr_cls.unsqueeze(1)
                )
                tuple_feat_base_domain_curr_cls = torch.split(
                    tensor_feat_base_domain_curr_cls, self.bs_match, dim=0
                )

                counter_curr_cls_base_domain = 0

                # tuple_feat_base_domain_curr_cls is a tuple of splitted part
                for feat_base_domain_curr_cls in tuple_feat_base_domain_curr_cls:
                    if flag_match_min_dist:  # if epoch > 0:flag_match_min_dist=True
                        # Need to compute over batches of
                        # feature due to device Memory out errors
                        # Else no need for loop over
                        # tuple_feat_base_domain_curr_cls;
                        # could have simply computed
                        # tensor_feat_curr_domain_curr_cls -
                        # tensor_feat_base_domain_curr_cls
                        dist_same_class_base_domain_curr_domain = torch.sum(
                            (
                                tensor_feat_curr_domain_curr_cls
                                - feat_base_domain_curr_cls
                            )
                            ** 2,
                            dim=2,
                        )
                        # tensor_feat_curr_domain_curr_cls.shape torch.Size([184, 512])
                        # feat_base_domain_curr_cls.shape torch.Size([64, 1, 512])
                        # (tensor_feat_curr_domain_curr_cls - feat_base_domain_curr_cls).shape:
                        # torch.Size([64, 184, 512])
                        # dist_same_class_base_domain_curr_domain.shape:
                        # torch.Size([64, 184]) is the per element distance of
                        # the cartesian product of feat_base_domain_curr_cls vs
                        # tensor_feat_curr_domain_curr_cls
                        match_ind_base_domain_curr_domain = torch.argmin(
                            dist_same_class_base_domain_curr_domain, dim=1
                        )
                        # the batch index of the neareast neighbors
                        # len(match_ind_base_domain_curr_domain)=64
                        # theoretically match_ind_base_domain_curr_domain can
                        # be a permutation of 0 to 183 though of size 64
                        # sort_val, sort_idx = \
                        # torch.sort(dist_same_class_base_domain_curr_domain, dim=1)
                        del dist_same_class_base_domain_curr_domain

                    #  feat_base_domain_curr_cls.shape torch.Size([64, 1, 512])
                    for idx in range(feat_base_domain_curr_cls.shape[0]):
                        # counter_curr_cls_base_domain =0 at initialization

                        # ## global_inds_base_domain_curr_cls pick out base
                        # domain class label y_c images
                        global_pos_base_domain_curr_cls = (
                            global_inds_base_domain_curr_cls[
                                counter_curr_cls_base_domain
                            ].item()
                        )
                        if curr_domain_ind == base_domain_idx:
                            ind_match_global_curr_domain_curr_cls = (
                                global_pos_base_domain_curr_cls
                            )
                        else:
                            if flag_match_min_dist:  # if epoch > 0:match_min_dist=True
                                ind_match_global_curr_domain_curr_cls = (
                                    global_inds_curr_domain_curr_cls[
                                        match_ind_base_domain_curr_domain[idx]
                                    ].item()
                                )
                            else:  # if epoch == 0
                                ind_match_global_curr_domain_curr_cls = (
                                    global_inds_curr_domain_curr_cls[
                                        counter_curr_cls_base_domain
                                        % size_curr_domain_curr_cls
                                    ].item()
                                )

                        self.dict_virtual_dset2each_domain[counter_ref_dset_size][
                            "data"
                        ][curr_domain_ind] = self.dict_domain_data[curr_domain_ind][
                            "data"
                        ][
                            ind_match_global_curr_domain_curr_cls
                        ]
                        self.dict_virtual_dset2each_domain[counter_ref_dset_size][
                            "label"
                        ][curr_domain_ind] = self.dict_domain_data[curr_domain_ind][
                            "label"
                        ][
                            ind_match_global_curr_domain_curr_cls
                        ]
                        # @FIXME: label initially were set to random continuous
                        # value, which is a technique to check if
                        # every data has been filled
                        counter_curr_cls_base_domain += 1
                        counter_ref_dset_size += 1

            if counter_ref_dset_size != self.virtual_ref_dset_size:
                logger = Logger.get_logger()
                logger.info(f"counter_ref_dset_size {counter_ref_dset_size}")
                logger.info(f"self.virtual_ref_dset_size {self.virtual_ref_dset_size}")
                logger.warning(
                    "counter_ref_dset_size not equal to self.virtual_ref_dset_size"
                )
                raise RuntimeError(
                    "counter_ref_dset_size not equal to self.virtual_ref_dset_size"
                )

        for key in self.dict_virtual_dset2each_domain.keys():
            if (
                self.dict_virtual_dset2each_domain[key]["label"].shape[0]
                != self.num_domains_tr
            ):
                raise RuntimeError(
                    "self.dict_virtual_dset2each_domain, \
                                   one key correspond to value tensor not \
                                   equal to number of training domains"
                )

        # Sanity Check: Ensure paired points have the same class label
        wrong_case = 0
        for key in self.dict_virtual_dset2each_domain.keys():
            for d_i in range(self.dict_virtual_dset2each_domain[key]["label"].shape[0]):
                for d_j in range(
                    self.dict_virtual_dset2each_domain[key]["label"].shape[0]
                ):
                    if d_j > d_i:
                        if (
                            self.dict_virtual_dset2each_domain[key]["label"][d_i]
                            != self.dict_virtual_dset2each_domain[key]["label"][d_j]
                        ):
                            # raise RuntimeError("the reference domain has 'rows' with inconsistent class labels")
                            wrong_case += 1
        logger = Logger.get_logger()
        logger.info(f"Total Label MisMatch across pairs: {wrong_case}")
        if wrong_case != 0:
            raise RuntimeError(
                "the reference domain has 'rows' with inconsistent class labels"
            )

        list_ref_domain_each_domain = []
        list_ref_domain_each_domain_label = []
        for ind_ref_domain_key in self.dict_virtual_dset2each_domain.keys():
            list_ref_domain_each_domain.append(
                self.dict_virtual_dset2each_domain[ind_ref_domain_key]["data"]
            )
            list_ref_domain_each_domain_label.append(
                self.dict_virtual_dset2each_domain[ind_ref_domain_key]["label"]
            )

        tensor_ref_domain_each_domain_x = torch.stack(list_ref_domain_each_domain)
        tensor_ref_domain_each_domain_label = torch.stack(
            list_ref_domain_each_domain_label
        )

        logger.info(
            f"{tensor_ref_domain_each_domain_x.shape} "
            f"{tensor_ref_domain_each_domain_label.shape}"
        )

        del self.dict_domain_data
        del self.dict_virtual_dset2each_domain
        return tensor_ref_domain_each_domain_x, tensor_ref_domain_each_domain_label



import torch


def mk_img(i_h, i_ch=3, batch_size=5):
    img = torch.rand(i_h, i_h)  # uniform distribution [0,1]
    # x = torch.clamp(x, 0, 1)
    img.unsqueeze_(0)
    img = img.repeat(i_ch, 1, 1)  # RGB image
    img.unsqueeze_(0)
    img = img.repeat(batch_size, 1, 1, 1)
    return img


def mk_rand_label_onehot(target_dim=10, batch_size=5):
    label_scalar = torch.randint(high=target_dim, size=(batch_size,))
    label_scalar2 = label_scalar.unsqueeze(1)
    label_zeros = torch.zeros(batch_size, target_dim)
    label_onehot = torch.scatter(
        input=label_zeros, dim=1, index=label_scalar2, value=1.0
    )
    return label_onehot


def mk_rand_xyd(ims, y_dim, d_dim, batch_size):
    imgs = mk_img(i_h=ims, batch_size=batch_size)
    ys = mk_rand_label_onehot(target_dim=y_dim, batch_size=batch_size)
    ds = mk_rand_label_onehot(target_dim=d_dim, batch_size=batch_size)
    return imgs, ys, ds



import functools
import inspect


def store_args(method):
    """Stores provided method args as instance attributes."""
    argspec = inspect.getfullargspec(method)
    defaults = {}
    if argspec.defaults is not None:
        defaults = dict(zip(argspec.args[-len(argspec.defaults) :], argspec.defaults))
    if argspec.kwonlydefaults is not None:
        defaults.update(argspec.kwonlydefaults)
    arg_names = argspec.args[1:]

    @functools.wraps(method)
    def wrapper(*positional_args, **keyword_args):
        self = positional_args[0]
        # Get default arg values
        args = defaults.copy()
        # Add provided arg values
        for name, value in zip(arg_names, positional_args[1:]):
            args[name] = value
        args.update(keyword_args)
        self.__dict__.update(args)
        return method(*positional_args, **keyword_args)

    return wrapper



"""
Jigen Model Similar to DANN model
"""
import warnings

from torch.nn import functional as F

from domainlab import g_str_cross_entropy_agg
from domainlab.dsets.utils_wrapdset_patches import WrapDsetPatches
from domainlab.models.a_model_classif import AModelClassif
from domainlab.models.model_dann import mk_dann


def mk_jigen(parent_class=AModelClassif):
    """
    Instantiate a JiGen model

    Details:
        The model is trained to solve two tasks:
        1. Standard image classification;
        2. Source images are decomposed into grids of patches, which are then permuted. The task
        is recovering the original image by predicting the right permutation of the patches;

        The (permuted) input data is first fed into a encoder neural network
        and then into the two classification networks.
        For more details, see:
        Carlucci, Fabio M., et al. "Domain generalization by solving jigsaw puzzles."
        Proceedings of the IEEE/CVF Conference on Computer Vision and Pattern Recognition. 2019.

    Args:
        parent_class (AModel, optional): Class object determining the task
        type. Defaults to AModelClassif.

    Returns:
        ModelJiGen: model inheriting from parent class

    Input Parameters:
        list_str_y: list of labels,
        list_str_d: list of domains,
        net_encoder: neural network (input: training data, standard and shuffled),
        net_classifier_class: neural network (input: output of net_encoder;
        output: label prediction),
        net_classifier_permutation: neural network (input: output of net_encoder;
        output: prediction of permutation index),
        coeff_reg: total_loss = img_class_loss + coeff_reg * perm_task_loss
        nperm: number of permutations to use, by default 31
        prob_permutation: probability of shuffling image tiles

    Usage:
        For a concrete example, see:
        https://github.com/marrlab/DomainLab/blob/master/tests/test_mk_exp_jigen.py
    """

    class_dann = mk_dann(parent_class)

    class ModelJiGen(class_dann):
        """
        Jigen Model Similar to DANN model
        """

        def __init__(
            self,
            list_str_y,
            net_encoder,
            net_classifier_class,
            net_classifier_permutation,
            coeff_reg,
            n_perm=31,
            prob_permutation=0.1,
            meta_info=None,
        ):
            super().__init__(
                list_str_y,
                list_d_tr=None,
                alpha=coeff_reg,
                net_encoder=net_encoder,
                net_classifier=net_classifier_class,
                net_discriminator=net_classifier_permutation,
            )
            self.net_encoder = net_encoder
            self.net_classifier_class = net_classifier_class
            self.net_classifier_permutation = net_classifier_permutation
            self.meta_info = meta_info
            self.n_perm = n_perm
            self.prob_perm = prob_permutation

        def dset_decoration_args_algo(self, args, ddset):
            """
            JiGen need to shuffle the tiles of the original image
            """
            if self.meta_info is not None:
                args.nperm = (
                    self.meta_info["nperm"] - 1
                    if "nperm" in self.meta_info
                    else args.nperm
                )
                args.pperm = (
                    self.meta_info["pperm"] if "pperm" in self.meta_info else args.pperm
                )
            nperm = self.n_perm
            pperm = self.prob_perm
            ddset_new = WrapDsetPatches(
                ddset,
                num_perms2classify=nperm,
                prob_no_perm=1 - pperm,
                grid_len=args.grid_len,
                ppath=args.jigen_ppath,
            )
            return ddset_new

        def _cal_reg_loss(self, tensor_x, tensor_y, tensor_d, others):
            """
            JiGen don't need domain label but a pre-defined permutation index
            to calculate regularization loss

            We don't know if tensor_x is an original/un-tile-shuffled image or
            a permutated image, which is the task of cal_reg_loss to classify
            which permutation has been used or no permutation has been used at
            all (which also has to be classified)
            """
            _ = tensor_y
            _ = tensor_d
            if isinstance(others, list) or isinstance(others, tuple):
                vec_perm_ind = others[0]
            else:
                vec_perm_ind = others
            # tensor_x can be either original image or tile-shuffled image
            feat = self.extract_semantic_feat(tensor_x)
            logits_which_permutation = self.net_classifier_permutation(feat)
            # _, batch_target_scalar = vec_perm_ind.max(dim=1)
            batch_target_scalar = vec_perm_ind
            batch_target_scalar = batch_target_scalar.to(tensor_x.device)
            loss_perm = F.cross_entropy(
                logits_which_permutation,
                batch_target_scalar,
                reduction=g_str_cross_entropy_agg,
            )
            return [loss_perm], [self.alpha]

    return ModelJiGen



from domainlab.compos.a_nn_builder import AbstractFeatExtractNNBuilderChainNode
from domainlab.compos.nn_zoo.nn_alex import Alex4DeepAll, AlexNetNoLastLayer


def mkNodeFeatExtractNNBuilderNameAlex(arg_name4net, arg_val):
    class NodeFeatExtractNNBuilderAlex(AbstractFeatExtractNNBuilderChainNode):
        """NodeFeatExtractNNBuilderAlex.
        Uniform interface to return AlexNet and other neural network as feature
        extractor from torchvision or external python file"""

        def init_business(
            self, dim_out, args, isize=None, remove_last_layer=False, flag_pretrain=True
        ):
            """
            initialize **and** return the heavy weight business
            object for doing the real job
            :param request: subclass can override request object
            to be string or function
            :return: the constructed service object
            i_size is not used at all in this class
            """
            if not remove_last_layer:
                self.net_feat_extract = Alex4DeepAll(flag_pretrain, dim_out)
            else:
                self.net_feat_extract = AlexNetNoLastLayer(flag_pretrain)
            return self.net_feat_extract

        def is_myjob(self, args):
            """is_myjob.
            :param args: command line arguments:
                "--nname": name of the torchvision model
            """
            arg_name = getattr(args, arg_name4net)
            return arg_name == arg_val

    return NodeFeatExtractNNBuilderAlex
